
##################################################################################
# This file is part of NNERO.
#
# Copyright (c) 2024, Gaétan Facchinetti
#
# NNERO is free software: you can redistribute it and/or modify it 
# under the terms of the GNU General Public License as published by 
# the Free Software Foundation, either version 3 of the License, or any 
# later version. NNERO is distributed in the hope that it will be useful, 
# but WITHOUT ANY WARRANTY; without even the implied warranty of 
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU 
# General Public License along with NNERO. 
# If not, see <https://www.gnu.org/licenses/>.
#
##################################################################################

##################
#
# Definition of constants
# Use this method rather than using astropy units for efficiency
#
##################

from types import SimpleNamespace

CST_EV_M_S_K = SimpleNamespace()
CST_MSOL_MPC = SimpleNamespace()
CST_NO_DIM   = SimpleNamespace()
CONVERSIONS  = SimpleNamespace()

CST_EV_M_S_K.mass_hydrogen = 0.93878299831e+9 # in eV
CST_EV_M_S_K.mass_proton   = 0.938272e+9 # in eV
CST_EV_M_S_K.mass_helium   = 3.728400e+9 # in eV
CST_EV_M_S_K.sigma_Thomson = 6.6524616e-29 # in m^2
CST_EV_M_S_K.c_light       = 299792458 # in m / s
CST_EV_M_S_K.k_Boltz       = 8.617343e-5 # Boltzmann constant in eV/K
CST_EV_M_S_K.T0            = 2.7255 # K
CST_EV_M_S_K.Tnu0          = 1.9454 # K
CST_EV_M_S_K.rho_c_over_h2 = 1.0598e+10 # eV / m^3

CST_MSOL_MPC.rho_c_over_h2 = 2.7754e+11 # in units of h^2 M_odot Mpc^{-3}

CST_NO_DIM.YHe  = 0.245
CST_NO_DIM.Neff = 3.044

CONVERSIONS.yr_to_s   = 3600 * 24 * 365.25
CONVERSIONS.km_to_mpc = 3.2407792896393e-20


##################################################################################
# This file is part of NNERO.
#
# Copyright (c) 2024, Gaétan Facchinetti
#
# NNERO is free software: you can redistribute it and/or modify it 
# under the terms of the GNU General Public License as published by 
# the Free Software Foundation, either version 3 of the License, or any 
# later version. NNERO is distributed in the hope that it will be useful, 
# but WITHOUT ANY WARRANTY; without even the implied warranty of 
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU 
# General Public License along with NNERO. 
# If not, see <https://www.gnu.org/licenses/>.
#
##################################################################################

##################
#
# General neural network framework
#
##################

import numpy as np
import torch

import os
from copy import deepcopy
from os.path import join

from .data import DataSet, MetaData, DataPartition, MP_KEY_CORRESPONDANCE


class NeuralNetwork(torch.nn.Module):
    """
    A class wrapping py:class:`torch.nn.Module` for neural network models

    Parameters
    ----------
    name: str
        name of the neural network

    Attributes
    ----------
    name: str
        the name of the model
    metadata: Metadata
        metadata on which the model is trained
    partition: DataPartition
        partitioning of the data on which the model is trained
    train_loss: np.ndarray
        1D array training loss for each training epoch
    valid_loss: np.ndarray
        1D array validation losses for each training epoch
    train_accuracy: np.ndarray
        1D array training accuracy for each training epoch
    valid_accuracy: np.ndarray
        1D array validation accuracy for each training epoch
    """


    def __init__(self, name: str) -> None:

        self._name: str = name

        self._metadata:  (MetaData | None)      = None
        self._partition: (DataPartition | None) = None

        self._train_loss     = np.zeros(0)
        self._valid_loss     = np.zeros(0)
        self._train_accuracy = np.zeros(0)
        self._valid_accuracy = np.zeros(0)

        self._struct         = np.empty(0)

        


    def save(self, path: str = ".", save_partition: bool = True) -> None:
        """
        Save the neural network model in a bunch of files.

        Parameters
        ----------
        path: str, optional
            path where to save the neural network
            -- default is the current directory "."
        save_partition: bool, optional
            if save_partition is false the partitioning of the data into
            train, valid and test is not saved (useless for instance once
            we have a fully trained model that we just want to use)
            -- default is True
        """

        # when partition is not required only print empty arrays
        if save_partition is False:
            partition       = deepcopy(self._partition)
            self._partition = DataPartition(np.zeros(0), np.zeros(0), np.zeros(0), np.zeros(0), np.zeros(0), np.zeros(0))

        # putting the model in eval mode
        self.eval()
        
        if len(self.struct) == 0:
            # saving the full class as a pickled object
            torch.save(self, join(path, self._name + ".pth"))
        else:
            # saving the state of the weights (recommended)
            torch.save(self._model.state_dict(), join(path, self._name + "_weights.pth"))

            # add extra information for the structure of the model
            with open(join(path, self._name + "_struct.npy"), 'wb') as file:
                np.save(file, self._struct, allow_pickle=False)

            # add extra information about the metadata used for training
            self.metadata.save(join(path, self._name + "_metadata"))
            
            # add extra information about the partition used for training
            self.partition.save(join(path, self._name + "_partition"))

            # add extra information about loss and accuracy during training
            with open(join(path, self._name + "_loss.npz"), 'wb') as file:
                np.savez(file, 
                         train_loss = self._train_loss,
                         train_accuracy = self._train_accuracy,
                         valid_loss = self._valid_loss,
                         valid_accuracy = self._valid_accuracy)

        # put the partition back to its original value
        if save_partition is False:
            self._partition = partition


    def load_weights_and_extras(self, path: str) -> None:
        """
        loads the network weights and extra information

        Parameters
        ----------
        path: str
            path to the network to load

        Raises
        ------
        ValueError
            If not all necessary files exists where path points.
        """
        
        # check if the needed files exist
        if os.path.isfile(path + '_weights.pth') and os.path.isfile(path + '_metadata.npz') :

            # set the weights of the model
            weights = torch.load(path  + '_weights.pth', weights_only=True)
            self._model.load_state_dict(weights)

            # fetch the metadata used during training
            self._metadata  = MetaData.load(path  + '_metadata')

            try:
                # fetch the partition used during training
                self._partition = DataPartition.load(path  + '_partition')
                
                # get the loss and accuracy obtained during training
                with open(path  + '_loss.npz', 'rb') as file:
                    data = np.load(file)
                    self._train_loss     = data.get('train_loss')
                    self._train_accuracy = data.get('train_accuracy')
                    self._valid_loss     = data.get('valid_loss')
                    self._valid_accuracy = data.get('valid_accuracy')
            except:
                # partition and saved loss are not necessary for the network
                # to work properly so we can simply pass if it does not work
                pass

            return None
        
        raise ValueError("Could not find a fully saved model at: " + path)


    def set_check_metadata_and_partition(self, dataset: DataSet, check_only: bool = False) -> None:
        """
        set and check the medatada and partition attributes
        
        Parameters
        ----------
        dataset: DataSet
            dataset to compare or to assign to the object
        check_only: bool, optional
            option to only compare the compatibility
            -- default is False
        
        Raises
        ------
        ValueError 
            if the dataset is incompatible with the current metadata or partition
        """
        
        # set and check the metadata
        if self.metadata is None and not check_only:
            self._metadata = dataset.metadata
        else:
            if self.metadata != dataset.metadata:
                raise ValueError("The metadata is incompatible with the previous round of training.")
        
        # set and check the partition
        if self.partition is None and not check_only:
            self._partition = dataset.partition
        else:
            if self.partition != dataset.partition:
                raise ValueError("The partition is incompatible with the previous round of training.")
            

    def print_structure(self):
        """
        prints the list of parameters in the model
        """
        
        total_params = 0
        print("| Parameters per layers:")
        print("| ----------------------")
        for name, parameter in self.named_parameters():
            if not parameter.requires_grad:
                continue
            params = parameter.numel()
            print('|', name, ':', params)
            total_params += params
        print("| ----------------------")
        print(f"| Total Trainable Params: {total_params}")
        print("  ----------------------")
            

    def info(self):
        
        print("#####################")
        print("Structure of the network")
        self.print_structure()
        print("#####################")
        print("Parameters:")
        inv_corres  = {value:key for key, value in MP_KEY_CORRESPONDANCE.items()}
        param_range = self.parameters_range
        other_name  = [inv_corres[param] for param in self.parameters_name]
        param_info = []
        max_param_len = 0
        for ip, param in enumerate(self.parameters_name):
            param_info.append(f"{param} ({str(other_name[ip])})")
            if len(param_info[-1]) > max_param_len:
                max_param_len = len(param_info[-1])

        for ip, param in enumerate(self.parameters_name):   
            print(f"-> {param_info[ip]:<{max_param_len}} in [{param_range[ip, 0]}, {param_range[ip, 1]}]")
        print("#####################")




    @property
    def name(self):
        return self._name

    @property
    def train_loss(self):
        return self._train_loss
    
    @property
    def valid_loss(self):
        return self._valid_loss
    
    @property
    def train_accuracy(self):
        return self._train_accuracy
    
    @property
    def valid_accuracy(self):
        return self._valid_accuracy
    
    @property
    def metadata(self):
        return self._metadata
    
    @property
    def partition(self):
        return self._partition
    
    @property
    def struct(self):
        return self._struct
    
    @property
    def parameters_name(self):
        return self._metadata.parameters_name
    
    @property
    def parameters_range(self):
        return np.vstack((self._metadata.parameters_min_val, self._metadata.parameters_max_val)).T



##################################################################################
# This file is part of NNERO.
#
# Copyright (c) 2024, Gaétan Facchinetti
#
# NNERO is free software: you can redistribute it and/or modify it 
# under the terms of the GNU General Public License as published by 
# the Free Software Foundation, either version 3 of the License, or any 
# later version. NNERO is distributed in the hope that it will be useful, 
# but WITHOUT ANY WARRANTY; without even the implied warranty of 
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU 
# General Public License along with NNERO. 
# If not, see <https://www.gnu.org/licenses/>.
#
##################################################################################

##################
#
# Definition of some functions to analyse chains and plot them
#
##################



import glob
import warnings
import os

import numpy as np
from copy import copy

from scipy.ndimage import gaussian_filter, gaussian_filter1d


from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Self

from .data         import MP_KEY_CORRESPONDANCE
from .predictor    import DEFAULT_VALUES
from .regressor    import Regressor
from .classifier   import Classifier
from .interpolator import Interpolator
from .predictor    import predict_Xe_numpy, predict_tau_numpy, predict_parameter_numpy


EMCEE_IMPORTED = False

try:
    import emcee
    EMCEE_IMPORTED = True
except:
    pass


def neutrino_masses(mnu_1, mnu_2 = 0.0, mnu_3 = 0.0, hierarchy = 'NORMAL'):

    ## DEFINE NEUTRINO MASSES BEFORE ANYTHING ELSE
    delta_m21_2 = 7.5e-5
    delta_m31_NO_2 = 2.55e-3
    delta_m31_IO_2 = 2.45e-3
    

    # degenerate neutrino masses (all equal to the first one)
    if hierarchy == 'DEGENERATE':
        mnu_2 = mnu_1
        mnu_3 = mnu_1

    # normal ordering
    if hierarchy == "NORMAL":
        mnu_2 = np.sqrt(mnu_1**2 + delta_m21_2) 
        mnu_3 = np.sqrt(mnu_1**2 + delta_m31_NO_2)

    # inverse ordering
    if hierarchy == "INVERSE":
        mnu_2 = np.sqrt(mnu_1**2 + delta_m31_IO_2) 
        mnu_3 = np.sqrt(mnu_1**2 + delta_m31_IO_2 + delta_m21_2)

    # create an array of neutrino masses
    m_neutrinos = np.array([mnu_1, mnu_2, mnu_3])
    return m_neutrinos


def to_CLASS_names(array: list[str] | np.ndarray):

    is_array = isinstance(array, np.ndarray)

    labels_correspondance = {value : key for key, value in MP_KEY_CORRESPONDANCE.items()}    
    array = [labels_correspondance[value] if value in labels_correspondance else value for value in array]

    if is_array is True:
        array = np.array(array)

    return array
    

def to_21cmFAST_names(array: list[str] | np.ndarray):

    is_array = isinstance(array, np.ndarray)
    array = [MP_KEY_CORRESPONDANCE[value] if value in MP_KEY_CORRESPONDANCE else value for value in array]

    if is_array is True:
        array = np.array(array)

    return array

#################################################
## CHAIN ANALYSIS TOOLS


class MPChain:
    """
    Class MPChain reading chains from MontePython output files

    Parameters
    ----------
    filename : str
        Path to the the file where the chain is stored.
    """

    def __init__(self, filename: str):
        
        self._filename: str = filename
        self.load()

    
    def load(self) -> None:
        """
        Load the chain and automatically remove the non-markovian points.
        """

        # read the file to get the chain
        with open(self._filename, 'r') as f:
            data = np.loadtxt(f) 

            self.weights = data[:, 0]
            self.lnlkl   = - data[:, 1]
            self._values = data[:, 2:].T

            self._total_n_steps = np.sum(self.weights, dtype=int)
            self._total_length  = self._values.shape[1]


        # reread the file to find the non markovian part of the chain
        with open(self._filename, 'r') as f:

            self._markov_index = 0
            for line in f:
                if (line.strip().startswith('#')) and ('update proposal' in line):
                    self._markov_index = int(line.split(' ')[2])

            # remove the non markovian part by default
            self._values = self._values[:, self._markov_index:]
            self.weights = self.weights[self._markov_index:]
            self.lnlkl = self.lnlkl[self._markov_index:]


        self._max_lnlkl = np.max(self.lnlkl)
        index_max_lnlkl = np.argmax(self.lnlkl)
        self._best_fit = self._values[:, index_max_lnlkl]

        self._n_params = self._values.shape[0]
        self._mean_value: np.ndarray = np.zeros(self._n_params)
        self._var_value:  np.ndarray = np.zeros(self._n_params)


    def remove_burnin(self, global_max_lnlkl: float) -> None:
        """
        Remove the burnin points according to the value of the maximum
        log likelihood over all chains. Only points of the chain that are 
        after its overcrossing of global_max_lnlkl - 3 are kept.

        Parameters
        ----------
        global_max_lnlkl : float
            Global maximum log likelihood over all chains
        """

        if np.all(self.lnlkl < (global_max_lnlkl - 3)):
            self._burnin_index = len(self.lnlkl)
        else:
            burnin_index = np.where(self.lnlkl >= global_max_lnlkl - 3)[0]

            if len(burnin_index) > 0:
                self._burnin_index = burnin_index[0]
            else:
                self._burnin_index = 0

        self._values  = self._values[:, self._burnin_index:]
        self.weights  = self.weights[self._burnin_index:]
        self.lnlkl    = self.lnlkl[self._burnin_index:]

        
    def values(self, discard: int = 0, thin: int = 1) -> np.ndarray:
        """
        Get the values of the chain.

        Parameters
        ----------
        discard : int, optional
            Number of initial points to discard, by default 0.
        thin : int, optional
            Thining factor (taking only one value every value of thin), by default 1.

        Returns
        -------
        np.ndarray with dimension (number of parameters, length of chain)
        """
        
        if discard > self._values.shape[-1]:
            discard = self._values.shape[-1]
            warnings.warn("All points in chain " + self._filename +  " discarded (discard >= chain length)")
        
        return self._values[:, discard::thin]
    
    def compute_stats(self) -> None:
        """
        Compute the mean and standard deviation within the chain.
        Should be called after `remove_burnin()`. 
        """

        n = np.sum(self.weights)

        self._mean_value = np.sum(self._values * self.weights[None, :], axis=-1) / n 
        self._var_value  = (np.sum((self._values**2) * self.weights[None, :], axis=-1) - n * (self._mean_value)**2) / (n-1)


    @property
    def markov_index(self):
        return self._markov_index

    @property
    def max_lnlkl(self):
        return self._max_lnlkl
    
    @property
    def burnin_index(self):
        return self._burnin_index
    
    @property
    def best_fit(self):
        return self._best_fit
    
    @property
    def mean_value(self):
        return self._mean_value
    
    @property
    def var_value(self):
        return self._var_value
    
    @property
    def length(self) -> int:
        """
        Number of accepted steps not counting burnin and non markovian points.

        Returns
        -------
        int
        """
        return self._values.shape[1]
    
    @property
    def total_length(self) -> int:
        """
        Total number of accepted steps

        Returns
        -------
        int
        """
        return self._total_length


    @property
    def total_n_steps(self) -> int:
        """
        Total number of steps.

        Returns
        -------
        int
        """
        return self._total_n_steps
    
    @property
    def n_steps(self) -> int:
        """
        Number of steps not counting burnin and non markovian points.

        Returns
        -------
        int
        """
        return np.sum(self.weights, dtype=int)
    
    @property
    def n_params(self):
        return self._n_params
    
    @property
    def acceptance_rate(self):
        return self.total_length/self.total_n_steps




class Samples(ABC):
    """
    Class containing all chains of a MCMC analysis
    
    Parameters
    ----------
    path : str
        Path to the chains.
    ids : list | np.ndarray | None, optional
        List of chains to take into accoung. If none all possible found chains are added. By default None.

    """

    def __init__(self, path : str, ids: list[int] | np.ndarray | None = None) -> None:
        
        self._path = path
        self._ids  = np.array(ids) if ids is not None else None

    @abstractmethod
    def flat(self, discard: np.ndarray | None = None, thin: None | int = None, **kwargs) -> np.ndarray:
        pass

    @property
    def path(self):
        return self._path
    
    @property
    def ids(self):
        return self._ids
    
    @property
    @abstractmethod
    def param_names(self) -> np.ndarray:
        pass

    @property
    @abstractmethod
    def scaling_factor(self) -> dict:
        pass


@dataclass
class GaussianInfo:

    mean: np.ndarray | None = None
    cov:  np.ndarray | None = None
    param_names: np.ndarray | list[str] | None = None

    def compatible_with(self, other: Self) -> bool:
        
        all_params = set(list(to_CLASS_names(self.param_names)) + list(to_CLASS_names(other.param_names)) )
       
        for param in all_params:
            if param in to_CLASS_names(self.param_names) and param in to_CLASS_names(other.param_names):
                index_1 = list(to_CLASS_names(self.param_names)).index(param)
                index_2 = list(to_CLASS_names(other.param_names)).index(param)

                if self.mean is not None and other.mean is not None:
                    if self.mean[index_1] != other.mean[index_2]:
                        return False 
                
        return True


class GaussianSamples(Samples):
    """
    Daughter class of Samples for gaussian generated chains.
    """

    def __init__(self, 
                 gaussians: list[GaussianInfo | str] | GaussianInfo | str,
                 add_tau: bool = False,
                 params: list[str] | None = None,
                 *, 
                 n: int = 200_000) -> None:

        super().__init__("", None)

        self._add_tau = add_tau

        # define a list of extra gaussian to add on top
        self._gaussians = gaussians
        if (not isinstance(self._gaussians, list)):
            self._gaussians = [self._gaussians]

        # if given string input read the data
        # and check for compatibility between mean values
        for ig in range(len(self._gaussians)):
            if isinstance(self._gaussians[ig], str):
                self._gaussians[ig] = self.load_data(self._gaussians[ig])
            
            if not self._gaussians[0].compatible_with(self._gaussians[ig]):
                raise ValueError('fiducial parameters should agree')

        # parameters on that we will be contained 
        # in the total covariance matrix
        self._params = params

        if self._params is None and self._gaussians is not None:
            self._params = self._gaussians[0].params

        self._params = to_CLASS_names(self._params)

        # get all the parameters in total
        all_params = []
        for g in self._gaussians:
            for p in g.param_names:
                if p not in all_params:
                    all_params.append(p)

        # build a big inverse covariance matrix
        # that contains all parameters in the input
        # covariance matrices
        all_n = len(all_params)
        all_inv_cov = np.zeros((all_n, all_n))
        all_mean    = np.zeros(all_n)

        for g in self._gaussians:
            this_params = list(to_CLASS_names(g.param_names))
            indices     = [all_params.index(param) for param in this_params]
            all_inv_cov[np.ix_(indices, indices)] += np.linalg.inv(g.cov)
            
            if g.mean is not None:
                all_mean[indices] = g.mean


        # select the part of the covariance matrix for the desired parameters
        all_cov        = np.linalg.inv(all_inv_cov)
        self._indices  = [all_params.index(param) for param in self._params]

        # construct a GaussianInfo object from the total cov and mean defined above
        self._gaussian       = GaussianInfo(all_mean[self._indices], all_cov[np.ix_(self._indices, self._indices)], self._params)
        self._generated_data = np.random.multivariate_normal(all_mean, all_cov, n).T

        self._all_params = all_params

        # add tau_ion as a parameter
        if self._add_tau is True:
            self._params = np.array(['tau_reio'] + list(self._params))
        


    def load_data(self, filename) -> GaussianInfo:

        gaussian = GaussianInfo()

        with open(filename, 'rb') as file:
            data = np.load(file)
            gaussian.cov         = data.get('cov', None)
            gaussian.param_names = list(to_CLASS_names(data['params'])) if 'params' in data else None
            gaussian.mean        = data.get('fiducial', None)

            if gaussian.mean is None:
                gaussian.mean = data.get('mean', None)

        return gaussian

        

    def flat(self, discard: np.ndarray | None = None, thin: None | int = None, **kwargs) -> np.ndarray:
        
        if discard is None:
            discard = 0

        if thin is None:
            thin = 1
        
        flat_chain = self._generated_data[:, discard::thin]

        if self._add_tau is True:

            classifier = kwargs.get('classifier', None)
            regressor  = kwargs.get('regressor', None)

            if regressor is None:
                regressor = Regressor.load()

            mask = np.full_like(flat_chain, fill_value=True, dtype=bool)

            for ip, param in enumerate(to_CLASS_names(regressor.parameters_name)):
                if param in self._all_params: # only take from index 1 as tau is the zeroth
                    index = list(self._all_params).index(param)
                    val_range = regressor.parameters_range
                    mask[index, :] =  flat_chain[index, :] < val_range[ip, 1]
                    mask[index, :] =  (flat_chain[index, :] > val_range[ip, 0]) & mask[index, :]

            valid_columns = np.all(mask, axis=0)
            flat_chain = flat_chain[:, valid_columns]

            tau = compute_tau(flat_chain, self._all_params, classifier, regressor)

            return np.vstack((tau[None, :], flat_chain[self._indices, :]))

        return flat_chain[self._indices, :]

    @property
    def param_names(self) -> np.ndarray:
        return self._params
    
    @property
    def all_param_names(self) -> np.ndarray:
        return self._all_params

    @property
    def scaling_factor(self) -> dict:
        return {name: 1.0 for name in self.param_names}
    
    @property
    def gaussian(self) -> GaussianInfo:
        return self._gaussian


class EMCEESamples(Samples):
    """
    Daughter class of Samples for emcee chains.
    """

    def __init__(self, path : str, add_tau: bool = False) -> None:
        
        if EMCEE_IMPORTED is False:
            print("Cannot import emcee, therefore cannot work with emcee chains")
            return None
        
        super().__init__(path, None)


        self._add_tau = add_tau
        self.load_chains()

        # Check for convergence criterion
        self._has_converged = False
        self._autocorr      = None
        
        try:
            self._autocorr = self._reader.get_autocorr_time()
            self._has_converged = True
        except emcee.autocorr.AutocorrError:
            self._autocorr = self._reader.get_autocorr_time(quiet = True)
            pass
        
        

    def load_chains(self):
        
        self._reader = emcee.backends.HDFBackend(self.path)
        
        filename = os.path.join(*self.path.split('/')[:-1], *self.path.split('/')[-1].split('.')[:-1])

        with open(filename + '.npz', 'rb') as file:

            data = np.load(file)
            self._parameters_theta = data.get('parameters_theta', None)
            self._parameters_xi    = data.get('parameters_xi', None)
            self._xi               = data.get('xi', None)

        if self._add_tau is False:
            self._param_names = np.array(list(self._parameters_theta) + list(self._parameters_xi))
        else:
            self._param_names = np.array(['tau_reio'] + list(self._parameters_theta) + list(self._parameters_xi))
        


    def flat(self, discard: np.ndarray | None = None, thin: None | int = None, **kwargs) -> np.ndarray:
        
        burnin_discard = int(np.max(2*self._autocorr)) if self._autocorr is not None else 0

        if discard is None:
            discard = 0

        if thin is None:
            thin = 1


        flat_chain = self._reader.get_chain(discard=burnin_discard + discard, thin=thin, flat = True)
        flat_chain = np.hstack((flat_chain, np.tile(self._xi, (flat_chain.shape[0], 1)))).T

        if self._add_tau is True:

            classifier = kwargs.get('classifier', None)
            regressor  = kwargs.get('regressor', None)
            tau = compute_tau(flat_chain, self.param_names, classifier, regressor)

            flat_chain = np.vstack((tau[None, :], flat_chain))
            
        return flat_chain

    

    def convergence(self):

        if self._has_converged is False:
            print("Not converged yet")
            print(self._autocorr)

        
    @property
    def reader(self):
        return self._reader
    
    @property
    def autocorr(self) -> np.ndarray:
        return self._autocorr

    @property
    def param_names(self) -> np.ndarray:
        return self._param_names
    
    @property
    def scaling_factor(self) -> dict:
        return {name: 1.0 for name in self.param_names}


def save_sampling_parameters(filename: str, 
                             parameters_theta : list[str], 
                             parameters_xi: list[str], 
                             xi: np.ndarray):
    with open(filename.split('.')[0] + '.npz', 'wb') as file:

        np.savez(file, 
                 parameters_theta=to_CLASS_names(parameters_theta), 
                 parameters_xi=to_CLASS_names(parameters_xi), 
                 xi=xi)


def compute_tau(flat_chain: np.ndarray, 
                param_names: list[str] | np.ndarray,
                classifier: Classifier | None = None,
                regressor: Regressor | None = None) -> np.ndarray:

      

        # get the ordered list of parameters
        if regressor is None:
            regressor  = Regressor.load()
        if classifier is None:
            classifier = Classifier.load()

        r_parameters = to_CLASS_names(regressor.metadata.parameters_name)
        parameters = to_CLASS_names(copy(param_names))

        if isinstance(parameters, list):
            parameters = np.array(parameters)

        if 'tau_reio' in parameters:
            parameters = parameters[parameters != 'tau_reio']

        for param in parameters:
            
            if param == 'sum_mnu':
                param = 'm_nu1'

            # make a list of the given parameters that are needed for the interpolator
            # in the order they are
            data_for_param = []
            for ip, param in enumerate(parameters):
                
                if param == 'sum_mnu':
                    parameters[ip] = 'm_nu1'

                if param in r_parameters:
                    data_for_param.append(param)
            
            # get the data sample 
            data = np.empty((len(r_parameters), flat_chain.shape[-1])) 

            # find the ordering in which data_sample is set in prepare_data_plot
            indices_to_plot = [list(parameters).index(param) for param in data_for_param]

        for ip, param in enumerate(r_parameters): 
            
            # if we ran the MCMC over that parameter
            if param in parameters[indices_to_plot]:
                index = list(parameters[indices_to_plot]).index(param)
                data[ip, :] = flat_chain[index, :]
            else:
                data[ip, :] = DEFAULT_VALUES[to_21cmFAST_names([param])[0]]

        #return data 
        tau = predict_tau_numpy(data.T, classifier, regressor)

        return tau




def compute_parameter(flat_chain: np.ndarray, 
                      param_names: list[str] | np.ndarray,
                      classifier: Classifier | None = None,
                      interpolator: Regressor | None = None,
                      parameter: str | None = None) -> np.ndarray:

        
        # get the ordered list of parameters
        if interpolator is None:
            if interpolator is not None:
                interpolator = Interpolator.load("DefaultInterpolator_" + parameter)
            else:
                raise ValueError("Need to pass the parameter to interpolate to the predictor.")
        
        if classifier is None:
            classifier = Classifier.load()

        # parameters that the interpolator needs
        i_parameters = to_CLASS_names(interpolator.metadata.parameters_name)

        # parameters that are given in the sample
        parameters = to_CLASS_names(copy(param_names))

        if isinstance(parameters, list):
            parameters = np.array(parameters)

        # if tau_reio is given in the parameters we remove it
        if 'tau_reio' in parameters:
            flat_chain = flat_chain[parameters != 'tau_reio'] 
            parameters = parameters[parameters != 'tau_reio']
            

        # make a list of the given parameters that are needed for the interpolator
        # in the order they are
        data_for_param = []
        for ip, param in enumerate(parameters):
            
            if param == 'sum_mnu':
                parameters[ip] = 'm_nu1'

            if param in i_parameters:
                data_for_param.append(param)
        
        # get the data sample 
        data = np.empty((len(i_parameters), flat_chain.shape[-1])) 

        # find the ordering in which data_sample is set in prepare_data_plot
        indices_to_plot = [list(parameters).index(param) for param in data_for_param]

        for ip, param in enumerate(i_parameters): 
            
            # if we ran the MCMC over that parameter
            if param in parameters[indices_to_plot]:
                index = list(parameters[indices_to_plot]).index(param)
                data[ip, :] = flat_chain[index, :]
            else:
                data[ip, :] = DEFAULT_VALUES[param]

        return predict_parameter_numpy(data.T, classifier, interpolator, parameter)




class MPSamples(Samples):
    """
    Daughter class of Samples for MontePython chains.
    """
 

    def __init__(self, 
                 path: str, 
                 ids: list[int] | np.ndarray | None = None):
        
        super().__init__(path, ids)
        self._chains : list[MPChain] = []
        
        self.load_chains()
        self.load_paramnames()

        self._max_lnlkl  = np.max([chain.max_lnlkl for chain in self._chains])
        chain_max_lnlkl  = np.argmax([chain.max_lnlkl for chain in self._chains])
        self._best_fit = self._chains[chain_max_lnlkl].best_fit

        for chain in self._chains:
            chain.remove_burnin(self._max_lnlkl)
        
        #######################
        # print some results

        max_markov_index = len(str(int(np.max([chain.markov_index for chain in self._chains]))))
        max_burnin_index = len(str(int(np.max([chain.burnin_index for chain in self._chains]))))
    
        for ic, chain in enumerate(self._chains):
            print(f'Chain {ic+1:<3} : Removed {chain.markov_index:<{max_markov_index}} non-markovian points, ' \
                  + f'{chain.burnin_index:<{max_burnin_index}} points of burn-in, keep ' + str(chain._values.shape[1]) \
                  + f' steps | (max_lnlkl = {chain.max_lnlkl:.2f}, acceptance_rate = {chain.acceptance_rate:.2f})' )
            
           
        #######################
        # compute some stats

        # define some global quantities (total number of steps and overall mean of the parameters)
        self._total_steps = np.sum(np.array([chain.n_steps for chain in self._chains]), dtype=int)

        self._total_mean = np.zeros(self.n_params)
        for chain in self._chains:
            if chain.length > 0:
                chain.compute_stats()

            self._total_mean  = self._total_mean + chain.n_steps * chain.mean_value
        self._total_mean = self._total_mean / self._total_steps

        self.load_scaling_factor()

           


    def load_chains(self) -> None:
        
        # look for chains in the folder
        chains_path = self.path +  '_*.txt'
        self._chains_name = np.array(glob.glob(chains_path))
        ids = np.array([int(name.split('.')[-2].split('_')[-1]) for name in self._chains_name], dtype=int)
        self._chains_name = self._chains_name[np.argsort(ids)]

        # raise an error if no chain is found
        if len(self._chains_name) == 0:
            raise ValueError("No chain found at " + chains_path)

        # redefine the chain name list from the given ids
        if self.ids is not None:
            self._chains_name = self._chains_name[self.ids]

        # define the number of chains
        self.n_chains = len(self._chains_name)
        
        # prepare an array for the non markovian chain
        self._markov = np.zeros(self.n_chains, dtype=int)

        # read all chains
        self._chains = [MPChain(filename) for filename in self._chains_name]

        self.n_params = self._chains[0].values().shape[0]
        

    def load_paramnames(self):
        
        self._names = np.empty(0)
        self._latex_names = np.empty(0)
       
        with open(self.path + '.paramnames', 'r') as f:
            for line in f:
                ls = line.split('\t')
                self._names = np.append(self._names, ls[0][:-1])
                self._latex_names = np.append(self._latex_names, r'${}$'.format(ls[1][1:-2]))


    def load_scaling_factor(self) -> None:

        self._scaling_factor = {}
        
        with open(os.path.join(*[*(self._path.split('/')[:-1]), 'log.param']), 'r') as file:
            for line in file:
                if (not line.strip().startswith('#')) and ('data.parameters' in line):
                    name  = line.split('[')[1].strip("[] ='")
                    value = float(line.split('[')[2].strip("[] ='\n").split(',')[-2].strip("[] ='"))
                    self._scaling_factor[name] = value  


    def flat(self, discard: np.ndarray | None = None, thin: None | int = None, **kwargs) -> np.ndarray:
        """
        Flatten samples of all chains.

        Parameters
        ----------
        discard : np.ndarray | None, optional
            Number of points to discard at begining of the chain, by default None (0).
        thin : None | int, optional
            Reduce the size of the sample by taking one point every `thin`, by default None.
            If Nont compute the reduced size such that the total length of the sample is 10000.

        Returns
        -------
        np.ndarray
            Sample in a 2 dimensional array of shape (# of parameters, # of points)
        """
        
        if isinstance(discard, int):
            discard = np.full(self.n_chains, discard)

        if discard is None:
            discard = np.zeros(self.n_chains, dtype=int)

        if thin is None:
            m_total_length = 0
            for ichain, chain in enumerate(self.chains):
                m_total_length = m_total_length + chain.values(discard = discard[ichain], thin=1).shape[1]

            if m_total_length > 1e+4:
                thin = int(m_total_length/10000)
            else:
                thin = 1

        res = np.empty((self.n_params, 0))
        for ichain, chain in enumerate(self.chains):
            res = np.concatenate((res, chain.values(discard = discard[ichain], thin=thin)), axis=-1)
        
        return res
    
    
    def convergence(self) -> np.ndarray:
        """
        Gelman-Rubin criterion weighted by the length of the chain as implemented
        in MontePython.

        Returns
        -------
        np.ndarray
            R-1 for all parameters
        """

        within  = np.zeros(self.n_params)
        between = np.zeros(self.n_params)
        
        for chain in self.chains :
            within  = within  + chain.n_steps * chain.var_value 
            between = between + chain.n_steps * (chain.mean_value - self.total_mean)**2

        within  = within / self.total_steps
        between = between / (self.total_steps-1)

        return between/within
    


    def covmat(self, 
               discard: np.ndarray | None = None, 
               params_in_cov : list[str] | None = None) -> np.ndarray:
        """
        Covariance matrix.

        Parameters
        ----------
        discard : np.ndarray | None, optional
            Number of points to discard at begining of the chain, by default None (0).
        data_to_cov : list[str] | None, optional
            List of parameters to put in the covariance matrix (in the order of that list).
            If None consider all parameters available.

        Returns
        -------
        np.ndarray
            Covariance matric (n, n) array
        """
        
        points = prepare_data_plot(self, params_in_cov, discard, thin=1)        
        return np.cov(points)




    def print_best_fit(self, discard: np.ndarray | None = None, **kwargs):
        
        samples_flat = self.flat(discard=discard, thin=1, **kwargs)
        med  = np.median(samples_flat, axis=1)
        mean = np.mean(samples_flat, axis=1)

        nc = np.zeros(len(self.param_names), dtype=int)
        for ip, param in enumerate(self.param_names):
            nc[ip] = len(param)

        max_nc = np.max(nc)

        for ip, param in enumerate(self.param_names):
            fill = " " * (max_nc - nc[ip])
            print(param +  fill + ' : \t' + str(self.best_fit[ip]) + " | " + str(med[ip]) + " | " + str(mean[ip]))

     


    @property
    def chains(self):
        return self._chains
        
    @property
    def max_lnlkl(self):
        return self._max_lnlkl  
    
    @property
    def best_fit(self):
        return self._best_fit
    
    @property
    def param_names(self):
        return self._names
    
    @property
    def param_names_latex(self):
        return self._latex_names
    
    @property
    def scaling_factor(self):
        return self._scaling_factor
    
    @property
    def total_steps(self):
        return self._total_steps
    
    @property
    def total_mean(self):
        return self._total_mean
    




def prepare_data_plot(samples: Samples, data_to_plot, discard = 0, thin = 1, **kwargs):

    data_to_plot = to_CLASS_names(data_to_plot)

    # give the transformation rules between the parameter if there is to be one
    def transform_data(data, name, param_names):

        if name == 'sum_mnu':
            if 'm_nu1' in param_names:
                index_mnu = data_to_plot.index('sum_mnu')
                data[index_mnu] = np.sum(neutrino_masses(data[index_mnu], hierarchy='NORMAL'), axis = 0)

                # change also the param_names
                param_names[index_mnu] = name
            else:
                raise ValueError('Impossible to obtain ' + name + ' from the input data sample')
            
        return param_names


    data = samples.flat(discard=discard, thin=thin, **kwargs)
    param_names = copy(samples.param_names)

    # rescaling the data according to the scaling factor
    for iname, name in enumerate(samples.param_names):
        if (samples.scaling_factor[name] != 1) and (name != '1/m_wdm' and name != 'm_wdm'): 
            # note that we keep the warm dark matter transformed
            # just means that we need to be carefull with the units
            data[iname] = samples.scaling_factor[name] * data[iname]


    # first transform the data if necessary
    if 'sum_mnu' in data_to_plot:
        param_names = transform_data(data, 'sum_mnu', param_names)


    # reduce the data to the indices to plot
    indices_to_plot = [np.where(param_names == param)[0][0] for param in data_to_plot if param in param_names]
    data = data[indices_to_plot]

    # remove outliers for tau_reio
    if 'tau_reio' in data_to_plot:
        index_tau_reio = data_to_plot.index('tau_reio')

        mask = data[index_tau_reio] < 0.1
        n_outliers = np.count_nonzero(~mask)
        
        if n_outliers > 0 : 
            warnings.warn(f'Removing {n_outliers} outlier points for tau_reio: ' + str(data[index_tau_reio, ~mask]))

        data = data[:, mask]
        

    return data




#################################################
## PLOTTING TOOLS

MATPLOTLIB_IMPORTED = False

try:

    import matplotlib.pyplot as plt
    import matplotlib.colors as mpc

    from dataclasses import dataclass

    MATPLOTLIB_IMPORTED = True    

    class AxesGrid:

        def __init__(self, 
                    n: int, 
                    *, 
                    scale: float = 2.0,
                    wspace: float = 0.05, 
                    hspace: float  = 0.05, 
                    labels: list[str] | None = None,
                    names: list[str] | None = None,
                    **kwargs):
            
            # close all pre-existing figures and disable interactive mode
            plt.close('all')
            
            # define a figure object and the grid spec
            self._fig = plt.figure(figsize=(scale*n, scale*n), constrained_layout=False)
            self._spec = self._fig.add_gridspec(ncols=n, nrows=n, wspace=wspace, hspace=hspace)

            # initialise the length
            self._size: int = n

            # initialise an empty list
            self._axs: list[plt.Axes] = [None for i in range(int(n*(n+1)/2))]

            # define the edges of the axes
            self._edges: np.ndarray = np.full((n, 2), fill_value=None)

            # define the labels of the axes
            self._labels: list[str] = [r'${{{}}}_{}$'.format(r'\theta', i) for i in range(self.size)] if labels is None else labels

            # define the name of the parameter attached to each axis
            self._names: list[str] = self._labels if names is None else names

            # define the titles of the axes (showing mean and quantiles if asked)
            self._titles: list[list[str]] = [[] for i in range(self.size)]

            # define the text objects holding the title
            self._titles_text : list[list[plt.Text | None]] = [[] for i in range(self.size)]  

            # define the axes on the grid
            for i in range(n):
                for j in range(i+1):

                    k = self.index_1D(i, j)
                    self._axs[k] = self._fig.add_subplot(self._spec[i, j])

                    if i < n-1:
                        self._axs[k].xaxis.set_ticklabels([])
                    if j > 0: 
                        self._axs[k].yaxis.set_ticklabels([])

                    self._axs[0].yaxis.set_ticklabels([])


            # define default font and rotation of ticks
            self._fontsize: float = self._axs[0].xaxis.label.get_size()
            self._ticks_rotation: float = 50
            self._titles_color: list[str] = [[] for i in range(self.size)]
            self._default_color: str = 'blue'

            self.update_labels(**kwargs)



        def get(self, i: int | str, j: int | str):

            i, j = (self.index_from_name(k) if isinstance(k, str) else k for k in [i, j])
            return self._axs[self.index_1D(i, j)]
        

        def index_1D(self, i: int, j: int):
            
            if j > i:
                raise ValueError("j must be less or equal than i") 

            return int(i*(i+1)/2 + j)
        
        def indices_2D(self, k):
        
            i = np.arange(0, k+2, dtype=int)
            ind_triangle = i*(i+1)/2
            row    = np.searchsorted(ind_triangle, k, side='right')-1
            column = int(k - ind_triangle[row])

            return row, column


        # show the plot
        def show(self):
            self._fig.show()


        # change one label name
        def set_label(self, i:int, name:str):
            self._labels[i] = name

            if i > 0:
                self.get(i, 0).set_ylabel(name)
            
            self.get(self.size-1, i).set_xlabel(name)


        # update all the label properties
        def update_labels(self, **kwargs):
            
            self._labels   = kwargs.get('labels', self._labels)
            self._fontsize = kwargs.get('fontsize', self._fontsize)
            self._ticks_rotation = kwargs.get('ticks_rotation', self._ticks_rotation)

            for i in range(1, self.size):
                k = self.index_1D(i, 0)
                self._axs[k].set_ylabel(self._labels[i], fontsize=self._fontsize)
                self._axs[k].tick_params(axis='y', labelsize=self._fontsize-2)
                for tick in self._axs[k].get_yticklabels():
                        tick.set_rotation(self._ticks_rotation)

            for j in range(self.size):
                k = self.index_1D(self.size-1, j)
                self._axs[k].set_xlabel(self._labels[j], fontsize=self._fontsize)
                self._axs[k].tick_params(axis='x', labelsize=self._fontsize-2)
                for tick in self._axs[k].get_xticklabels():
                    tick.set_rotation(self._ticks_rotation)

        #update the edges
        def update_edges(self, axis: int | str, min: float, max: float):
        
            j = self.index_from_name(axis) if isinstance(axis, str) else axis
            self._edges[j, :] = np.array([min, max]) 

            for i in range(j, self.size):
                self.get(i, j).set_xlim([self._edges[j, 0], self._edges[j, -1]])

            for k in range(0, j):
                self.get(j, k).set_ylim([self._edges[j, 0], self._edges[j, -1]])
        

        # update the titles properties
        def add_title(self, axis: int, new_titles: str, color: str | None = None):
            
            self._titles_color[axis].append(color if color is not None else self._default_color)
            self._titles[axis].append(new_titles)
            self._titles_text[axis].append(None)

        def update_titles(self, height = 1.05, spacing = 1.9, fontsize = None):

            if fontsize is None:
                fontsize = self._fontsize

            for j in range(self.size):
                k = self.index_1D(j, j)
                for it, title in enumerate(self._titles[j]):
                    total_height = height if it == 0 else height+it*spacing*1e-2*fontsize
                    if self._titles_text[j][it] is None: 
                        self._titles_text[j][it] = self._axs[k].text(0.5, total_height, title, fontsize=fontsize, color = self._titles_color[j][it], ha='center', transform=self._axs[k].transAxes)
                    else:
                        self._titles_text[j][it].set_position((0.5, total_height))
                        self._titles_text[j][it].set_text(title)
                        self._titles_text[j][it].set_fontsize(fontsize)

        def index_from_name(self, name: str | list[str]):
            
            if isinstance(name, str):
                return self.names.index(name)
            else:
                return [self.names.index(na) for na in name]

        # List of properties

        @property
        def fig(self):
            return self._fig
        
        @property
        def spec(self):
            return self._spec
        
        @property
        def size(self):
            return self._size
        
        @property
        def edges(self):
            return self._edges
        
        @edges.setter
        def edges(self, value: np.ndarray):
            self._edges = value

        @property
        def labels(self):
            return self._labels
        
        @labels.setter
        def labels(self, value: list[str]):
            self._labels = value
            self.update_labels()

        @property
        def names(self):
            return self._names


    
    @dataclass
    class ProcessedData:

        hists_1D  : np.ndarray | None = None
        hists_2D  : np.ndarray | None = None
        edges     : np.ndarray | None = None
        centers   : np.ndarray | None = None
        levels    : np.ndarray | None = None
        q         : np.ndarray | None = None
        mean      : np.ndarray | None = None
        median    : np.ndarray | None = None
        bestfit   : np.ndarray | None = None
        quantiles : np.ndarray | None = None
        samples   : np.ndarray | None = None
        size      : int | None = None



    def compute_quantiles(sample: np.ndarray, q: float, bins: int | np.ndarray = 30) -> tuple[np.ndarray, np.ndarray]:
        """
        Give the q-th quantile of the input sample.

        Parameters
        ----------
        sample : np.ndarray
            1D array of data points.
        q : float
            Quantile value.
        bins : int | np.ndarray, optional
            Binning of the histogram that is used 
            for a first approximation of the quantile 
            edges, by default 30.

        Returns
        -------
        tuple[np.ndarray, np.ndarray]
        min,max bounds
        """

        def split_in_groups(lst):
        
            res = [[int(lst[0])]]
            i = 0

            for il, element in enumerate(lst):
                element = int(element)
                if il > 0:
                    if element - lst[il-1] == 1:
                        res[i].append(element)
                    else:
                        res.append([element])
                        i = i+1

            return res
            
        # define the histogram from the sample
        hist, edges = np.histogram(sample, bins=bins, density=True)
        hist_count, _ = np.histogram(sample, bins=bins)

        # normalise the histogram to 1
        hist = hist/np.max(hist)

        # define the total sum of the histogram
        s_hist = np.sum(hist)

        # order the histogram values
        ordered_hist  = np.sort(hist)[::-1]
        order_of_hist = np.argsort(hist)[::-1]

        # compute the cumulative sum and get all indices that
        # contribute to the quantile
        sum_hist = np.cumsum(ordered_hist)/s_hist 
        indices = np.arange(0, len(hist), dtype=int)[sum_hist < q]

        # the first bin is already high enough we return its boundaries
        if len(indices) == 0:

            e_min = np.empty((1, 2))
            e_max = np.empty((1, 2))

            index_one = order_of_hist[0]

            e_min[0, 0] = edges[:-1][index_one]
            e_max[0, 0] = edges[:-1][index_one]
            e_min[0, 1] = edges[1:][index_one]
            e_max[0, 1] = edges[1:][index_one]

            ilg = 0
            total_count_small = 0

            x_min = np.zeros(1)
            x_max = np.zeros(1)

            m = np.median(sample)

        else:

            # add an extra index to be sure we capture all the good bins
            final_index = int(indices[-1]+1)

            # if we have same size hist values where sum_hist = q then
            # we need to consider all possible bins that can contribute
            # we put all these indices into a table
            multiple_indices = [final_index]

            while final_index < len(ordered_hist)-1 and (ordered_hist[final_index] == ordered_hist[final_index+1]):
                multiple_indices.append(final_index+1)
                final_index = final_index+1

            # chack for each last index what is the one that minimize the number of groups 
            len_groups = []

            for index in multiple_indices:
                bin_indices = np.sort(order_of_hist[np.concatenate((indices, np.array([index])))]) # all bins contributing
                len_groups.append(len(split_in_groups(bin_indices)))

            index_min_groups = np.argmin(len_groups)
            bin_indices    = np.sort(order_of_hist[np.concatenate((indices, np.array([multiple_indices[index_min_groups]])))])
            groups = split_in_groups(bin_indices)


            # here size is different from length,
            # size refers to the 'physical' size
            size_groups   = []
            count_groups  = []
            sample_groups = []

            # get what is inside the groups
            # approximate values of the other groups
            e_min = np.empty((len(groups), 2))
            e_max = np.empty((len(groups), 2))

            x_min = np.empty(len(groups))
            x_max = np.empty(len(groups))

            for ig, group in enumerate(groups):
                index_min, index_max = np.min(group), np.max(group)
                
                e_max[ig, 0], e_max[ig, 1] = edges[:-1][index_max], edges[1:][index_max]
                e_min[ig, 0], e_min[ig, 1] = edges[:-1][index_min], edges[1:][index_min]

                size_groups.append(float(e_max[ig, 1]-e_min[ig, 0]))
                count_groups.append(int(np.sum(hist_count[group])))

                mask = (sample > e_min[ig, 0]) & (sample < e_max[ig, 1])
                sample_groups.append(sample[mask])

                x_min[ig] = e_min[ig, 0]
                x_max[ig] = e_max[ig, 1]

                # adjust the exact value for the largest group
                # and keep only the bins of the smaller groups


            ilg = np.argmax(size_groups)
            total_count_small = np.sum(count_groups) - count_groups[ilg]
            m = np.median(sample_groups[ilg]) if len(groups) > 0 else np.median(sample)


        # define the grid for the finner search
        x_m = np.linspace(e_min[ilg, 0], e_min[ilg, 1], 30)
        x_p = np.linspace(e_max[ilg, 0], e_max[ilg, 1], 30)


        # for each value of x_m find the value of x_p such that quantile is q
        mask  = (sample >= x_m[:, None, None]) & (sample <=  x_p[None, :, None])
        count = (np.count_nonzero(mask, axis=-1) + total_count_small)/len(sample)
        m_x_p = x_p[np.argmin(np.abs(count - q), axis=1)]
        
        if count[0, -2] < q and count[1, -1] < q and count[1, -2] < q:
            x_min[ilg], x_max[ilg] = x_m[0], x_p[-1]
            return x_min, x_max

        # remove the values where x_p is just equal to the max
        # at the very edege of the grid 
        mask_xp = np.argmin(m_x_p != x_p[-1])

        # if we have removed all points then problem
        # we actually do not mask, it is just that we
        # cannot really find the q contour with our
        # resolution. Return the max value
        if mask_xp == 0:
            x_min[ilg] = x_m[-1]
            x_max[ilg] = x_p[-1]
        else:
            xm  = x_m[:mask_xp+1]
            xp  = m_x_p[:mask_xp+1]

            index = np.argmin(np.sqrt((2*m - xm - xp)**2))
            x_min[ilg], x_max[ilg] = xm[index], xp[index]

        return x_min, x_max





    def generate_contours(samples: np.ndarray, 
                        bins: int = 20, 
                        q = [0.68, 0.95],
                        smooth_2D: bool = False,
                        smooth_1D: bool = False,
                        sigma_smooth: float = 1.5,) -> ProcessedData:

        data = ProcessedData()

        n = samples.shape[0]

        hists_1D   = np.empty((n, bins))
        hists_2D   = np.empty((n, n, bins, bins))    # 2D array of 2D array
        levels     = np.empty((n, n, len(q)+1)) # 2D array of 1D array

        quantiles  = np.full((len(q), n, 2), fill_value=np.nan)

        q = np.array(q)
        if np.any(q != sorted(q)):
            raise ValueError('levels should be given in ascending order')
            # check that the input levels are in ascending order

        # edges and centers of the histograms
        edges   = np.vstack([np.linspace(np.min(s), np.max(s), bins+1) for s in samples])
        centers = (edges[:, :-1] + edges[:, 1:]) / 2

        mean   = np.mean(samples, axis=-1)
        median = np.median(samples, axis=-1)


        # loop over all places with 1D histograms
        for i in range(n):
            
            hists_1D[i, :], _ = np.histogram(samples[i, :], bins=edges[i, :], density=True)
            hists_1D[i] = gaussian_filter1d(hists_1D[i], sigma=sigma_smooth) if smooth_1D is True else hists_1D[i]
            
            hists_1D[i] = hists_1D[i] / np.max(hists_1D[i])

            # evaluate the quantiles
            for il, q_val in enumerate(q):
                try:
                    x_min, x_max = compute_quantiles(samples[i, :], q_val, bins=bins)
                except:
                    warnings.warn(f"Impossible to compute quantiles for entry {i}")

                if len(x_min) > 1 or len(x_max) > 1:
                    warnings.warn(f"Quantiles not given for what appears to be a multimodal distribution {i}")
                else:
                    try:
                        quantiles[il, i, 0], quantiles[il, i, 1] = x_min[0], x_max[0]
                    except Exception as e:
                        print(quantiles[il, i, 0], quantiles[il, i, 1], x_min, x_max)
                        raise e

        
        # loop over all places with 2D histograms
        for i in range(1, n):
            for j in range(i):

                hists_2D[i, j, :, :], _, _ = np.histogram2d(samples[j, :], samples[i, :], bins=[edges[j, :], edges[i, :]], density=True)
                hists_2D[i, j] = gaussian_filter(hists_2D[i, j], sigma=sigma_smooth) if smooth_2D is True else hists_2D[i, j]
                
                # Flatten the histogram to sort for cumulative density
                sorted_hist = np.sort(hists_2D[i, j].ravel())[::-1]  # Sort in descending order

                # Compute cumulative density
                cumulative = np.cumsum(sorted_hist) / np.sum(sorted_hist)

                # Find threshold levels for the desired confidence intervals
                levels[i, j, :-1] = sorted_hist[np.searchsorted(cumulative, q[::-1])]
                levels[i, j, -1]  = 1.1*np.max(sorted_hist)

        data.hists_2D  = hists_2D
        data.hists_1D  = hists_1D
        data.samples   = samples
        data.edges     = edges
        data.centers   = centers
        data.levels    = levels
        data.mean      = mean
        data.median    = median
        data.quantiles = quantiles
        data.size      = hists_1D.shape[0]
        data.q         = q
            
        return data



    def plot_data(grid: AxesGrid, 
                data : ProcessedData, 
                show_hist: bool    = False, 
                show_surface: bool = True, 
                show_contour: bool = False,
                show_quantiles: list[bool] = [False, False],
                show_mean: bool = False,
                show_title: bool = True,
                show_points: bool = False,
                show_marginals: bool = True,
                redefine_edges: bool = True,
                q_in_title: int = 0.68,
                colors: list[str]  = 'orange',
                axes : list[int] | np.ndarray | None = None,
                exclude_quantiles : int | str | list[int] | list[str] = [],
                exclude_mean : int | str | list[int] | list[str] = [],
                exclude_title : int | str | list[int] | list[str] = [],
                exclude_marginal : int | str | list[int] | list[str]  = [],
                alpha_surface: list[float] = 1.0,
                alpha_contour: list[float] = 1.0,
                alpha: list[float] | None = None, 
                linestyle_contour: list[str] = ['-'],
                linewidth_contour: list[float] = [1.0]):
        
        # preparing plotting option arrays
        linewidth_contour, alpha_surface, alpha_contour, colors, exclude_quantiles, exclude_mean, exclude_title = ([array] if isinstance(array, float) else array for array in [linewidth_contour, alpha_surface, alpha_contour, colors, exclude_quantiles, exclude_mean, exclude_title])
        linestyle_contour = [linestyle_contour] if isinstance(linestyle_contour, str) else linestyle_contour
        exclude_quantiles, exclude_mean, exclude_title, exclude_marginal = (grid.index_from_name(exclude) if (len(exclude) > 0 and isinstance(exclude[0], str)) else exclude for exclude in [exclude_quantiles, exclude_mean, exclude_title, exclude_marginal])


        if axes is None:
            axes = np.arange(0, data.size)

        # we we just give the alphas, overrides the separate values
        # for contours and surfaces
        if alpha is not None:
            alpha = [alpha] if isinstance(alpha, float) else alpha
            alpha_surface = alpha
            alpha_contour = alpha


        # first define the colors we will need to use
        contour_colors = [mpc.to_rgba(color, alpha_contour[ic]) if isinstance(color, str) else color for ic, color in enumerate(colors)]
        surface_colors = [mpc.to_rgba(color, alpha_surface[ic]) if isinstance(color, str) else color for ic, color in enumerate(colors)]

        # if we provide one color and we ask for more levels then
        # we define new colors automatically
        if len(contour_colors) == 1:
            
            pastelness = np.array([0.7]) if len(data.levels[0, 0]) == 3 else np.linspace(0.5, 0.8, len(data.levels[0, 0])-2)
            pastelness = pastelness[:, None] * np.ones((1, 4))
            pastelness[:, -1] = 0

            # add custom pastel colors to the stack of colors
            contour_colors = np.vstack(((1.0 - pastelness) * np.array(contour_colors) + pastelness, contour_colors))
            surface_colors = np.vstack(((1.0 - pastelness) * np.array(surface_colors) + pastelness, surface_colors))
            
        # plot the contours and points
        for i in range(1, data.size):
            for j in range(i):

                hist = data.hists_2D[i, j]

                if show_points is True:

                    # first thin the samples so that we plot only 5000 points
                    n = data.samples.shape[-1]
                    r = np.max([1, int(n/5000)])

                    grid.get(axes[i], axes[j]).scatter(data.samples[j, ::r], data.samples[i, ::r], marker='o', edgecolors='none', color = contour_colors[-1], s=2, alpha=0.5)


                if show_hist is True:
                    extent = [data.edges[j, 0], data.edges[j, -1], data.edges[i, 0], data.edges[i, -1]]
                    grid.get(axes[i], axes[j]).imshow(hist.T, origin='lower', extent=extent, cmap='Greys', aspect='auto')

                
                if show_surface is True:
                    try:
                        grid.get(axes[i], axes[j]).contourf(*np.meshgrid(data.centers[j], data.centers[i]), hist.T, levels=data.levels[i, j], colors=surface_colors)
                    except ValueError as e:
                        print("Error for axis : ", i, j)
                        raise e

                if show_contour is True:
                    grid.get(axes[i], axes[j]).contour(*np.meshgrid(data.centers[j], data.centers[i]), hist.T, levels=data.levels[i, j], colors=contour_colors, linestyles=linestyle_contour, linewidths=linewidth_contour, zorder=10)


        
        # fill in the 1D histograms
       
        for i in range(0, data.size):
            
            if show_marginals and (i not in exclude_marginal):
                grid.get(axes[i], axes[i]).stairs(data.hists_1D[i], edges=data.edges[i, :], color = colors[0], linestyle = linestyle_contour[0], linewidth = linewidth_contour[0])

            if (show_mean is True) and (i not in exclude_mean):
                grid.get(axes[i], axes[i]).axvline(data.mean[i], color=contour_colors[0], linewidth=0.5, linestyle='--')


            for iq, quantile in enumerate(data.quantiles[::-1]):

                if (show_quantiles[iq] is True) and (i not in exclude_quantiles):

                    # fill the histogram in terms of the first quantile
                    mask_edges = (data.edges[i, :] <= quantile[i, 1]) & (data.edges[i, :] >= quantile[i, 0])

                    # corresponding mask for the histogram values
                    mask_hist  = (mask_edges[:-1]*mask_edges[1:] == 1)

                    if len(data.hists_1D[i, mask_hist]) > 0:
                        # plot the quantiles with a shaded histogram
                        grid.get(axes[i], axes[i]).stairs(data.hists_1D[i, mask_hist], edges=data.edges[i, mask_edges], color = contour_colors[iq], fill=True, alpha=0.5)
                        
            title_color = contour_colors[1]
            title_color[-1] = 1.0

            # get the number of quantiles
            if q_in_title in data.q:
                jq = np.where(q_in_title == data.q)[0][0]
            else:
                raise ValueError('quantile in title should be a q value given in generate_contour')
            
            if (show_title is True) and (i not in exclude_title):
                if not np.isnan(data.quantiles[jq, i, 0]) and not np.isnan(data.quantiles[jq, i, 1]) :
                    grid.add_title(axes[i], r'${:.3g}$'.format(data.median[i], color=colors[0]) + '$^{{ +{:.2g} }}_{{ -{:.2g} }}$'.format(data.quantiles[jq, i, 1] - data.median[i], data.median[i] - data.quantiles[jq, i, 0] ), color=title_color)  
                else:
                    grid.add_title(axes[i], r'${:.3g}$'.format(data.median[i], color=colors[0]), color=title_color) 
        
        
        grid.update_titles()


        # (re)define the grid edges
        if redefine_edges is True:
            
            new_boundaries = data.edges[:, [0, -1]]

            old_boundaries = grid.edges[axes]
            old_boundaries[old_boundaries[:, 0] == None] = data.edges[old_boundaries[:, 0] == None, :][:, [0, -1]]

            new_min = np.minimum(new_boundaries[:, 0], old_boundaries[:, 0]) 
            new_max = np.maximum(new_boundaries[:, 1], old_boundaries[:, 1])

            new_edges = np.vstack((new_min, new_max)).T

            grid.edges[axes, :] = new_edges 
            
            for j in range(0, data.size):
            
                for i in range(j, data.size):
                    grid.get(axes[i], axes[j]).set_xlim([grid.edges[axes[j], 0], grid.edges[axes[j], -1]])
                    
                    if i > j:
                        grid.get(axes[i], axes[j]).set_ylim([grid.edges[axes[i], 0], grid.edges[axes[i], -1]])




    def plot_2D_marginal(ax: plt.Axes,
                        data : ProcessedData, 
                        i: int,
                        j: int,
                        show_hist: bool    = False, 
                        show_surface: bool = True, 
                        show_contour: bool = False,
                        show_points: bool  = False,
                        colors: list[str]  = 'orange', 
                        alphas: list[float] = 1.0):
        


        if i < j:
            i, j = j, i

        hist = data.hists_2D[i, j].T
    

        alphas, colors = ([array] if isinstance(array, float) else array for array in [alphas, colors])

        # first define the colors we will need to use
        contour_colors = [mpc.to_rgba(color, alphas[ic]) if isinstance(color, str) else color for ic, color in enumerate(colors)]

        # if we provide one color and we ask for more levels then
        # we define new colors automatically colors
        if len(contour_colors) == 1:
            
            pastelness = np.array([0.7]) if len(data.levels[0, 0]) == 3 else np.linspace(0.5, 0.8, len(data.levels[0, 0])-2)
            pastelness = pastelness[:, None] * np.ones((1, 4))
            pastelness[:, -1] = 0

            # add custom pastel colors to the stack of colors
            contour_colors = np.vstack(((1.0 - pastelness) * np.array(contour_colors) + pastelness, contour_colors))
        
        if show_points is True:

            # first thin the samples so that we plot only 5000 points
            n = data.samples.shape[-1]
            r = np.max([1, int(n/10000)])

            ax.scatter(data.samples[j, ::r], data.samples[i, ::r], marker='o', edgecolors='none', color = contour_colors[-1], s=2, alpha=0.5)

        if show_hist is True:
            extent = [data.edges[j, 0], data.edges[j, -1], data.edges[i, 0], data.edges[i, -1]]
            ax.imshow(hist, origin='lower', extent=extent, cmap='Greys', aspect='auto')

                
        if show_surface is True:
            try:
                ax.contourf(*np.meshgrid(data.centers[j], data.centers[i]), hist, levels=data.levels[i, j], colors=contour_colors)
            except ValueError as e:
                print("Error for axis : ", i, j)
                raise e

        if show_contour is True:
            ax.contour(*np.meshgrid(data.centers[j], data.centers[i]), hist, levels=data.levels[i, j], colors=contour_colors)

        



    def prepare_data_Xe(samples: Samples, 
                    data_to_plot: list[str] | np.ndarray,  
                    discard: int = 0, 
                    thin: int = 100,
                    *,
                    classifier: Classifier | None = None,
                    regressor: Regressor | None = None,):
        
        data_for_Xe = []

        data_to_plot = to_CLASS_names(data_to_plot)

        # get the ordered list of parameters
        if regressor is None:
            regressor  = Regressor.load()

        if classifier is None:
            classifier = Classifier.load()

        parameters = regressor.metadata.parameters_name

        for param in data_to_plot:
            
            if param == 'sum_mnu':
                param = 'm_nu1'

            if (param in MP_KEY_CORRESPONDANCE) and (MP_KEY_CORRESPONDANCE[param] in parameters):
                data_for_Xe.append(param)

        
        labels_correspondance = {value : key for key, value in MP_KEY_CORRESPONDANCE.items()}
        
        # get the data sample 
        data_sample = prepare_data_plot(samples, data_for_Xe, discard=discard, thin=thin, regressor = regressor, classifier = classifier)
        data = np.empty((len(parameters), data_sample.shape[-1])) 

        # find the ordering in which data_sample is set in prepare_data_plot
        indices_to_plot = [np.where(samples.param_names == param)[0][0] for param in data_for_Xe if param in samples.param_names]

        for ip, param in enumerate(parameters): 
            
            # if we ran the MCMC over that parameter
            if labels_correspondance[param] in samples.param_names[indices_to_plot]:
                index = list(samples.param_names[indices_to_plot]).index(labels_correspondance[param])
                data[ip] = data_sample[index, :]
            else:
                data[ip, :] = DEFAULT_VALUES[param]

        return data
        


    def get_Xe_stats(samples: Samples, 
                    data_to_plot: list[str] | np.ndarray,  
                    nbins: int = 100, 
                    discard: int = 0, 
                    thin: int = 100,
                    *,
                    classifier: Classifier | None = None,
                    regressor: Regressor | None = None,
                    smooth: bool = False,
                    sigma_smooth: float = 1.5,
                    **kwargs):

        data = prepare_data_Xe(samples, data_to_plot, discard, thin, classifier=classifier, regressor=regressor).T

        for kw, val in kwargs.items():
            if kw in regressor.parameters_name:
                ikw = list(regressor.parameters_name).index(kw)
                data[:, ikw] = val
            else:
                raise ValueError("Need to pass a kwarg that is in the parameter list of the classifier / regressor")

        Xe = predict_Xe_numpy(theta=data, classifier=classifier, regressor=regressor)

        # here remove some outliers that should not have 
        # passed the likelihood condition
        if np.count_nonzero(Xe[:, -1]==-1)/len(Xe) > 0.01:
            warnings.warn("More than 1 percent of outliers with late reionization")

        Xe = Xe[Xe[:, -1] > 0]
        z = regressor.metadata.z
        
        mean, med = np.mean(Xe, axis=0), np.median(Xe, axis=0)
        min, max  = np.min(Xe), np.max(Xe)

        log_bins = np.linspace(np.log10(min), np.log10(max), nbins)
        log_hist = np.zeros((len(z), len(log_bins)-1))

        lin_bins = np.linspace(min, max, nbins)
        lin_hist = np.zeros((len(z), len(lin_bins)-1))

        # make an histogram for each value of z
        for iz, x in enumerate(Xe.T):

            log_hist[iz], _ = np.histogram(np.log10(x), bins=log_bins, density = True)
            log_hist[iz] = gaussian_filter1d(log_hist[iz], sigma=sigma_smooth) if smooth is True else log_hist[iz]
            log_hist[iz] = log_hist[iz]/np.max(log_hist[iz])
            
            lin_hist[iz], _ = np.histogram(x, bins=lin_bins, density = True)
            lin_hist[iz] = gaussian_filter1d(lin_hist[iz], sigma=sigma_smooth) if smooth is True else lin_hist[iz]
            lin_hist[iz] = lin_hist[iz]/np.max(lin_hist[iz])
        
        log_bins = 10**((log_bins[:-1] + log_bins[1:])/2.0)
        lin_bins = (lin_bins[:-1] + lin_bins[1:])/2.0
        
        return z, mean, med, log_hist, log_bins, lin_hist, lin_bins




    def get_Xe_tanh_stats(samples: Samples, 
                            nbins: int = 100, 
                            discard: int = 0, 
                            thin : int = 100, 
                            x_inf: float = 2e-4, 
                            *,
                            smooth: bool = False,
                            sigma_smooth: float = 1.5,
                            **kwargs):

        def Xe_class(z, z_reio = 8.0):
            return 0.5 * ( 1+np.tanh(  (1+z_reio)/(1.5*0.5)*(1-((1+z)/(1+z_reio))**(1.5))   ) )   

        index = list(samples.param_names).index('z_reio')
        z_reio = samples.flat(discard=discard, thin=thin, **kwargs)[index, :]

        z = np.linspace(0, 35, 100)
        Xe = Xe_class(z, z_reio[:, None]) + x_inf

        mean, med = np.mean(Xe, axis=0), np.median(Xe, axis=0)
        min, max  = np.min(Xe), np.max(Xe)

        log_bins = np.linspace(np.log10(min), np.log10(max), nbins)
        log_hist = np.zeros((len(z), len(log_bins)-1))

        lin_bins = np.linspace(min, max, nbins)
        lin_hist = np.zeros((len(z), len(lin_bins)-1))

        # make an histogram for each value of z
        for iz, x in enumerate(Xe.T):

            log_hist[iz], _ = np.histogram(np.log10(x), bins=log_bins, density = True)
            log_hist[iz] = gaussian_filter1d(log_hist[iz], sigma=sigma_smooth) if smooth is True else log_hist[iz]
            log_hist[iz] = log_hist[iz]/np.max(log_hist[iz])
            
            lin_hist[iz], _ = np.histogram(x, bins=lin_bins, density = True)
            lin_hist[iz] = gaussian_filter1d(lin_hist[iz], sigma=sigma_smooth) if smooth is True else lin_hist[iz]
            lin_hist[iz] = lin_hist[iz]/np.max(lin_hist[iz])
        
        log_bins = 10**((log_bins[:-1] + log_bins[1:])/2.0)
        lin_bins = (lin_bins[:-1] + lin_bins[1:])/2.0
        
        return z, mean, med, log_hist, log_bins, lin_hist, lin_bins
    
except:
    pass



##################################################################################
# This file is part of NNERO.
#
# Copyright (c) 2024, Gaétan Facchinetti
#
# NNERO is free software: you can redistribute it and/or modify it 
# under the terms of the GNU General Public License as published by 
# the Free Software Foundation, either version 3 of the License, or any 
# later version. NNERO is distributed in the hope that it will be useful, 
# but WITHOUT ANY WARRANTY; without even the implied warranty of 
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU 
# General Public License along with NNERO. 
# If not, see <https://www.gnu.org/licenses/>.
#
##################################################################################

##################
#
# Definition of simple astrophysical quantities
# for the computation of UV-luminosity functions
#
##################

import numpy as np

from .cosmology import convert_array, dn_dm
from .constants import CONVERSIONS


def m_halo(hz:         float | np.ndarray,
           m_uv:       float | np.ndarray, 
           alpha_star: float | np.ndarray, 
           t_star:     float | np.ndarray, 
           f_star10:   float | np.ndarray,
           omega_b:    float | np.ndarray,
           omega_m:    float | np.ndarray) -> tuple[np.ndarray,  np.ndarray]:
    """
    Halo mass in term of the UV magnitude for a given astrophysical model

    Parameters
    ----------
    hz: float, np.ndarray 
        Shape (q, r). Hubble factor given in s^{-1}
    m_uv: float, np.ndarry (s,) or (r, s)
        UV magnitude.
    omega_b: float, np.ndaray (q,)
        Reduced abundance of baryons

    Returns
    -------
    numpy.ndarray with shape (q, r, s)
    """

    hz         = convert_array(hz)
    m_uv       = convert_array(m_uv)
    alpha_star = convert_array(alpha_star)[:, None, None]
    t_star     = convert_array(t_star)[:, None, None]
    f_star10   = convert_array(f_star10)[:, None, None]
    omega_b    = convert_array(omega_b)
    omega_m    = convert_array(omega_m)

    if len(hz.shape) == 1:
        hz = hz[None, :]

    hz = hz[..., None] # shape (q, r, 1)

    if len(m_uv.shape) == 1:
        m_uv = m_uv[None, None, :]

    if len(m_uv.shape) == 2:
        m_uv = m_uv[None, :, :]


    q: int = f_star10.shape[0]
    r: int = hz.shape[1]
    s: int = m_uv.shape[2]

    
    # conversion constant between star formation rate and magnitude
    gamma_UV = 1.15e-28 * 10**(0.4*51.63) / CONVERSIONS.yr_to_s
    
    # fraction of baryons over total matter
    fb = (omega_b/omega_m)[:, None, None]

    # define the return array and fill it with zeros
    res = np.zeros((q, r, s)) # shape (q, r, s)

    # mh_below are those for which f_star10 (m/1e+10) <= 1 (and f_star = f_b f_star10 (m/1e+10))
    # mh_below is of shape (q, r, s)
    mh_below = 1e+10 * ( gamma_UV/(hz*1e+10) * t_star / f_star10 / fb *  10**(-0.4*m_uv) )**(1.0/(alpha_star+1.0))

    # make a difference if f_star10 (m/1e+10) <= 1 or not
    # mask is of shape (q, r, s)
    mask = ((f_star10 * (mh_below/1e+10)**alpha_star) <= 1)

    # vals below are those for which f_star10 (m/1e+10) > 1  (and f_star = f_b)
    # mh_above is of shape (q, r, s)
    mh_above = gamma_UV / hz * t_star / fb *  10**(-0.4*m_uv) 

    # fill the return array with the correct values according to the mask
    res[mask]  = mh_below[mask]
    res[~mask] = mh_above[~mask]

    # check that all mh below statisfy the criterion (as it should)
    if np.any( ( np.tile(f_star10, (1, r, s))[~mask] * (res[~mask]/1e+10)**( np.tile(alpha_star, (1, r, s))[~mask])   ) <= 1):
        raise ValueError("Halo mass value (index below) should satisfy f_star10 * (mh_below/1e+10)**alpha_star < 1")

    # the mask is True when we are below i.e. f_star10 (m/1e+10)^alpha_star < 1, and False when we are above
    return res, mask



def dmhalo_dmuv(hz: float | np.ndarray,
             m_uv: float | np.ndarray, 
             alpha_star: float | np.ndarray, 
             t_star: float | np.ndarray, 
             f_star10: float | np.ndarray,
             omega_b: float | np.ndarray,
             omega_m: float | np.ndarray,
             *,
             mh: np.ndarray | None = None,
             mask: np.ndarray | None = None):
    

    if (mh is None) or (mask is None):
        mh, mask = m_halo(hz, m_uv, alpha_star, t_star, f_star10, omega_b, omega_m) # shape (q, r, s)

    alpha_star = convert_array(alpha_star)
    
    res        = - mh * np.log(10.0) * 0.4/(alpha_star[:, None, None]+1) # if f_star10 (m/1e+10)^alpha_star <= 1
    res[~mask] = - mh[~mask] * np.log(10.0) * 0.4 # if f_star10 (m/1e+10)^alpha_star > 1
    
    return res 
    


def f_duty(mh : float | np.ndarray, m_turn: float | np.ndarray):

    # result of shape (q, r, s)

    mh = convert_array(mh)
    m_turn = convert_array(m_turn)[:, None, None]

    return np.exp(-m_turn/mh)


def phi_uv(z:          float | np.ndarray,
           hz:         float | np.ndarray,
           m_uv:       float | np.ndarray,
           k:          np.ndarray,
           pk:         np.ndarray, 
           alpha_star: float | np.ndarray, 
           t_star:     float | np.ndarray, 
           f_star10:   float | np.ndarray,
           m_turn:     float | np.ndarray,
           omega_b:    float | np.ndarray,
           omega_m:    float | np.ndarray,
           h:          float | np.ndarray,
           sheth_a:    float = 0.322,
           sheth_q:    float = 1.0,
           sheth_p:    float = 0.3,
           *, 
           window: str = 'sharpk', 
           c: float = 2.5,
           mh: np.ndarray | None = None,
           mask: np.ndarray | None = None,
           dndmh: np.ndarray | None = None):
    
    """
    UV flux in Mpc^{-3}

    Parameters
    ----------
    m_uv: float, np.ndarry (s,)
    omega_b: float, np.ndaray (q,)

    Returns
    -------
    result of shape (q, r, s)
    """

    
    # result of shape (n, r, q) in Mpc^(-3)

    if (mh is None) or (mask is None):
        mh, mask = m_halo(hz, m_uv, alpha_star, t_star, f_star10, omega_b, omega_m)               # shape (q, r, s)
    
    dmh_dmuv = dmhalo_dmuv(hz, m_uv, alpha_star, t_star, f_star10, omega_b, omega_m, mh = mh, mask=mask) # shape (q, r, s)
    
    if dndmh is None:
        dndmh = dn_dm(z, mh, k, pk, omega_m, h, sheth_a=sheth_a, sheth_q=sheth_q, sheth_p=sheth_p, window=window, c=c) # shape (q, r, s)

    return f_duty(mh, m_turn) * dndmh * np.abs(dmh_dmuv)




##################################################################################
# This file is part of NNERO.
#
# Copyright (c) 2024, Gaétan Facchinetti
#
# NNERO is free software: you can redistribute it and/or modify it 
# under the terms of the GNU General Public License as published by 
# the Free Software Foundation, either version 3 of the License, or any 
# later version. NNERO is distributed in the hope that it will be useful, 
# but WITHOUT ANY WARRANTY; without even the implied warranty of 
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU 
# General Public License along with NNERO. 
# If not, see <https://www.gnu.org/licenses/>.
#
##################################################################################


import numpy as np
import torch
import torch.nn as nn

from .data      import TorchDataset, DataSet
from .network   import NeuralNetwork

import os
import pkg_resources

DATA_PATH = pkg_resources.resource_filename('nnero', 'nn_data/')


class Interpolator(NeuralNetwork):
    """
    Daughter class of :py:class:`NeuralNetwork` specialised for interpolators.
    
    Parameters
    ----------
    model: torch.nn.Module | None
        If not None, the model that will be used for classifier. 
        Otherwise, a new model is constructed from `n_input`, `n_hidden_features` 
        and `n_hidden_layers`. Default is None.
    dataset: Dataset
        Dataset on which the model will be trained. 
    n_input: int, optional
        Number of input on the neural network 
        (corresponds to the number of parameters).
        Default is 16.
    n_input: int, optional
        Number of output on the neural network 
        (corresponds to the number of redshift bins).
        Default is 50. Overriden if `dataset` is specified.
    n_hidden_features: int, optional
        Number of hidden features per layer. Default is 80.
    n_hidden_layers: int, optional
        Number of layers. Default is 5.
    name: str | None, optional
        Name of the neural network. If None, automatically set to DefaultClassifier.
        Default is None.
    parameter: str | None, optional 
        Parameter which we want to interpolate. Default is None, which only consider the first
        parameter in the extras of the dataset.
        
    Attributes
    ----------
    name: str
        the name of the model

    Raises
    ------
    ValueError
        When the parameter that the network needs to learn is not in the dataset. 
    """

    def __init__(self,
                 dataset: DataSet | None = None,
                 *, 
                 n_input: int = 15,
                 n_hidden_features: int = 80, 
                 n_hidden_layers: int = 5, 
                 model = None, 
                 parameter: str | None = None,
                 id_param: int | None = None,
                 name: str | None = None):

        
        if dataset is not None:

            n_input = len(dataset.metadata.parameters_name)

            if (parameter is not None) and (parameter not in dataset.extras_name):
                raise ValueError("Need to pass a parameter that is in the dataset.")
            
            if len(dataset.extras_name) == 0:
                raise ValueError("No extra parameters in the dataset.")
            
            self._id_param  = list(dataset.extras_name).index(parameter) if parameter is not None else 0
        
        else:

            self._id_param = id_param

         
        self._parameter = parameter

        if name is None:
            name = "DefaultInterpolator_" + parameter


        # give a default empty array for the structure
        # stays None if a complex model is passed as input
        struct = np.empty(0)

        # definition of the model if none is given as input
        if model is None:
            
            hidden_layers = []
            for _ in range(n_hidden_layers):
                hidden_layers.append(nn.Linear(n_hidden_features, n_hidden_features))
                hidden_layers.append(nn.ReLU())

            # create a sequential model
            model = nn.Sequential(nn.Linear(n_input, n_hidden_features), *hidden_layers, nn.Linear(n_hidden_features, 1))
        
            # save the structure of this sequential model and more
            struct = np.array([n_input, n_hidden_features, n_hidden_layers, self.parameter, self.id_param])
        
        # call the (grand)parent constructors
        super(Interpolator, self).__init__(name)
        super(NeuralNetwork, self).__init__()

        # structure of the model
        self._struct = struct
        
        # define the model
        self._model = model

        # set the dataset of the network
        if dataset is not None:
            self.set_check_metadata_and_partition(dataset)

    
    @property
    def id_param(self) -> int:
        return self._id_param
    
    @property
    def parameter(self) -> str:
        return self._parameter


    @classmethod
    def load(cls, path: str | None = None):
        """
        Loads an interpolator.

        Parameters
        ----------
        path: str | None
            Path to the saved files containing the interpolator data.
            If None automatically fetch the DefaultInterpolator.

        Returns
        -------
        Inteprolator
        """

        if path is None:
            path = os.path.join(DATA_PATH, "DefaultInterpolator")

        name = path.split('/')[-1]
        
        if os.path.isfile(path  + '_struct.npy'):

            with open(path  + '_struct.npy', 'rb') as file:
                struct  = np.load(file)

                if len(struct) == 5:

                    interpolator = Interpolator(n_input=int(struct[0]), 
                                          n_hidden_features=int(struct[1]), 
                                          n_hidden_layers=int(struct[2]),
                                          parameter=str(struct[3]),
                                          id_param=int(struct[4]),
                                          name=name)
                    interpolator.load_weights_and_extras(path)
                    interpolator.eval()

                    print('Model ' + str(name) + ' sucessfully loaded')

                    return interpolator
        
        # if the struct read is not of the right size
        # check for a pickled save of the full class
        # (although this is not recommended)
        if os.path.isfile(path  + '.pth') :
            interpolator = torch.load(path + ".pth")
            interpolator.eval()

            print('Model ' + str(name) + ' sucessfully loaded from a .pth archive')
            
            return interpolator
        
        raise ValueError("Could not find a fully saved interpolator model at: " + path)


        
    def forward(self, x):
        """
        Forward evaluation of the model.

        Parameters
        ----------
        x: torch.Tensor
            input features
        """

        return torch.flatten(torch.clamp(self._model(x), max=1))
    
    def loss(self, pred, target):
        return torch.mean(torch.abs(1.0 - torch.div(pred, target)))
    

    def test(self, dataset:DataSet) -> tuple[np.ndarray, np.ndarray]:
        """
        Test the efficiency of the regressor to reconstruct the parameter.

        Parameters
        ----------
        dataset : DataSet
            DataSet containing the training partition and the test partition.

        Returns
        -------
        tuple(np.ndarray, np.ndarray)
            Prediction for X_e and true values.
        """

        self.set_check_metadata_and_partition(dataset, check_only = True)

        # clean the dataset (# may have some issued for some values)
        # the dataset needs to be checked before hand
        y_test = dataset.extras_array[0, dataset.partition.early_test]
        x_test = dataset.x_array[dataset.partition.early_test]

        mask = ~(np.isnan(y_test) | np.isinf(y_test))
        x_test = torch.tensor(x_test[mask], dtype=torch.float32)
        y_test = torch.tensor(y_test[mask], dtype=torch.float32)
        
        with torch.no_grad():
            y_pred = self.forward(x_test)

        return y_pred.numpy(), y_test.numpy()
    

    

def train_interpolator(model: Interpolator, 
                    dataset: DataSet, 
                    optimizer: torch.optim.Optimizer, 
                    *, 
                    epochs = 50, 
                    learning_rate = 1e-3, 
                    verbose = True, 
                    batch_size = 64, 
                    **kwargs):
    """
    Trains a given regressor.

    Parameters
    ----------
    model : Regressor
        Regressor model to train.
    dataset : DataSet
        Dataset on which to train the regressor.
    optimizer : torch.optim.Optimizer
        Optimizer used for training.
    epochs : int, optional
        Number of epochs, by default 50.
    learning_rate : float, optional
        Learning rate for training, by default 1e-3.
    verbose : bool, optional
        If true, outputs a summary of the losses at each epoch, by default True.
    batch_size : int, optional
        Size of the training batches, by default 64.
    """

    # set or check the metadata and parition attributes of the model
    model.set_check_metadata_and_partition(dataset)

    # clean the dataset (# may have some issued for some values)
    # the dataset needs to be checked before hand
    y_train = dataset.extras_array[model.id_param, dataset.partition.selection_train]
    x_train = dataset.x_array[dataset.partition.selection_train]

    y_valid = dataset.extras_array[model.id_param, dataset.partition.selection_valid]
    x_valid = dataset.x_array[dataset.partition.selection_valid]

    mask = ~(np.isnan(y_train) | np.isinf(y_train))
    x_train = x_train[mask]
    y_train = y_train[mask]
    
    mask = ~(np.isnan(y_valid) | np.isinf(y_valid))
    x_valid = x_valid[mask]
    y_valid = y_valid[mask]
    

    # format the data for the regressor
    train_dataset = TorchDataset(x_train, y_train)
    valid_dataset = TorchDataset(x_valid, y_valid)
    train_loader  = torch.utils.data.DataLoader(train_dataset, batch_size=batch_size, shuffle=True, **kwargs)
    valid_loader  = torch.utils.data.DataLoader(valid_dataset, batch_size=batch_size, shuffle=False, **kwargs)

    # we have only one param_group here
    # we modify the learning rate of that group
    optimizer.param_groups[0]['lr'] = learning_rate

    # start loop on the epochs
    for epoch in range(epochs):
        
        train_loss     = np.array([])
        valid_loss     = np.array([])
        train_accuracy = np.array([])
        valid_accuracy = np.array([])

        # training mode
        model.train()
        
        for batch in train_loader:
            
            x_batch, y_batch = batch

            optimizer.zero_grad()
            y_pred = model.forward(x_batch)         
            loss   = model.loss(y_pred, y_batch)

            loss.backward()
            optimizer.step()
            
            train_loss     = np.append(train_loss, loss.item())
            train_accuracy = np.append(train_accuracy, 1-loss.item())


        # evaluation mode
        model.eval()
    
        with torch.no_grad():
            
            for batch in valid_loader:
                x_batch, y_batch = batch
            
                y_pred = model.forward(x_batch)        
                loss   = model.loss(y_pred, y_batch)
                
                valid_loss     = np.append(valid_loss, loss.item())
                valid_accuracy = np.append(valid_accuracy, 1-loss.item())
        
        # get the mean of all batches
        model._train_loss     = np.append(model._train_loss, np.mean(train_loss))
        model._valid_loss     = np.append(model._valid_loss, np.mean(valid_loss))
        model._train_accuracy = np.append(model._train_accuracy, np.mean(train_accuracy))
        model._valid_accuracy = np.append(model._valid_accuracy, np.mean(valid_accuracy))

        if verbose:
            print(f'Epoch [{epoch+1}/{epochs}], loss: ({model.train_loss[-1]:.4f}, {model.valid_loss[-1]:.4f}), accuracy = ({model.train_accuracy[-1]:.4f}, {model.valid_accuracy[-1]:.4f})')   



##################################################################################
# This file is part of NNERO.
#
# Copyright (c) 2024, Gaétan Facchinetti
#
# NNERO is free software: you can redistribute it and/or modify it 
# under the terms of the GNU General Public License as published by 
# the Free Software Foundation, either version 3 of the License, or any 
# later version. NNERO is distributed in the hope that it will be useful, 
# but WITHOUT ANY WARRANTY; without even the implied warranty of 
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU 
# General Public License along with NNERO. 
# If not, see <https://www.gnu.org/licenses/>.
#
##################################################################################

from .cosmology    import (ShortPowerSpectrumRange, optical_depth_no_rad, h_factor_no_rad, n_baryons, n_hydrogen)
from .constants    import (CST_MSOL_MPC, CST_EV_M_S_K, CST_NO_DIM, CONVERSIONS)
from .astrophysics import (phi_uv)
from .data         import (DataSet, MetaData, DataPartition, latex_labels, uniform_to_true, true_to_uniform)
from .network      import (NeuralNetwork)
from .classifier   import (Classifier, train_classifier)
from .regressor    import (Regressor, train_regressor)
from .interpolator import (Interpolator, train_interpolator)
from .predictor    import (input_values, predict_classifier, predict_xHII, predict_tau, predict_tau_from_xHII,
                         predict_classifier_numpy, predict_xHII_numpy, predict_tau_numpy, predict_tau_from_xHII_numpy,
                         predict_Xe, predict_Xe_numpy, predict_tau_from_Xe, predict_tau_from_Xe_numpy,
                         predict_parameter, predict_parameter_numpy, predict_interpolator, predict_interpolator_numpy)
from .mcmc         import (log_prior, log_likelihood, log_probability, initialise_walkers, Likelihood, UVLFLikelihood)
from .analysis     import (MPChain, Samples, MPSamples, EMCEESamples, GaussianInfo, GaussianSamples, neutrino_masses)

from .analysis import MATPLOTLIB_IMPORTED

if MATPLOTLIB_IMPORTED:
    from .analysis import(AxesGrid, plot_data,  generate_contours, ProcessedData, 
                          compute_quantiles, prepare_data_plot,
                          get_Xe_stats, get_Xe_tanh_stats)




##################################################################################
# This file is part of NNERO.
#
# Copyright (c) 2024, Gaétan Facchinetti
#
# NNERO is free software: you can redistribute it and/or modify it 
# under the terms of the GNU General Public License as published by 
# the Free Software Foundation, either version 3 of the License, or any 
# later version. NNERO is distributed in the hope that it will be useful, 
# but WITHOUT ANY WARRANTY; without even the implied warranty of 
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU 
# General Public License along with NNERO. 
# If not, see <https://www.gnu.org/licenses/>.
#
##################################################################################


import numpy as np
import torch

from .data         import MetaData, true_to_uniform
from .cosmology    import optical_depth_no_rad
from .classifier   import Classifier
from .regressor    import Regressor
from .interpolator import Interpolator
from .data         import MP_KEY_CORRESPONDANCE


DEFAULT_VALUES = {'F_STAR10' : -1.5, 'ALPHA_STAR' : 0.5, 't_STAR' : 0.5, 'F_ESC10' : -1.0, 'ALPHA_ESC' : 0.3, 'M_TURN' : 8.7,
            'Omdmh2' : 0.11933, 'Ombh2' : 0.02242,  'hlittle' : 0.6736, 'Ln_1010_As' : 3.047, 'POWER_INDEX' : 0.9665, 
            'INVERSE_M_WDM' : 0.0, 'NEUTRINO_MASS_1' : 0.0, 'FRAC_WDM' : 0.0, 'M_WDM' : '20.0', 'L_X' : 40.0, 'NU_X_THRESH' : 500,
            'LOG10_PMF_SB' : -5.0, 'PMF_NB' : -2.0}



## Note that due to a strange naming convention in 21cmFAST, Omch2 actually corresponded to omega_dm
## This notation is deprecated today, prefer to use Omdmh2

# ---------------------------------------------------
# CHECKS AND PREPARATION OF THE DATA TO FED TO THE NN

def check_values(vals, metadata : MetaData):

    params_name = metadata.parameters_name
    min_params  = metadata.parameters_min_val
    max_params  = metadata.parameters_max_val

    if len(vals.shape) == 1:
        vals = vals[None, :]
    elif len(params_name) != vals.shape[-1]:
        raise ValueError("The input parameter array should have last dimension of size " + str(len(params_name)))

    # error handling, check that inputs are in the correct range
    if not (np.all(vals >= min_params) and np.all(max_params >= vals)):
        
        id_min_problem, pos_min_problem = np.where(vals < min_params)
        id_max_problem, pos_max_problem = np.where(max_params < vals)

        out_str = "Some parameters input are not in the correct range:\n"
        for i in range(len(id_min_problem)):
            out_str = out_str + str(id_min_problem[i]) +  " -> " +  params_name[pos_min_problem[i]] + " : " + str(vals[id_min_problem[i], pos_min_problem[i]]) + " < min_trained_value = " + str(min_params[pos_min_problem[i]]) + "\n"
        for i in range(len(id_max_problem)):
            out_str = out_str + str(id_max_problem[i]) +  " -> " +  params_name[pos_max_problem[i]] + " : " + str(vals[id_max_problem[i], pos_max_problem[i]]) + " > max_trained_value = " + str(max_params[pos_max_problem[i]]) + "\n"

        out_str = out_str.strip('\n')
        raise ValueError(out_str)


    
def input_values(metadata: MetaData, default: str = DEFAULT_VALUES, **kwargs):

    params_name = metadata.parameters_name

    # all parameters the neural network have been trained on
    iparams = {value: index for index, value in enumerate(params_name)}

    # predefined default values for most common parameters
    vals = np.array([default[p] for p in params_name])
        
    # check that the arguments passed in kwargs were trained on
    kw_keys = np.array(list(kwargs.keys()))
    kw_vals = np.array(list(kwargs.values()))

    # translate the keys that could be comming with a different naming convention
    kw_keys = np.array([MP_KEY_CORRESPONDANCE[key] if key in MP_KEY_CORRESPONDANCE.keys() else key for key in kw_keys])

    # error handling, check that inputs are in the trained parameters list
    # concatenate params_name and kw_keys and get unique input, if all goes well
    # the resulting array should have the same length as params_name
    if len(np.unique(np.concatenate((params_name, kw_keys)))) != len(params_name):
        raise ValueError("Some arguments of " + str(kw_keys) + " are not in the trained parameters list: " + str(params_name))

    # give their value to the parameters
    vals[[iparams[kw] for kw in kw_keys]] = kw_vals

    # error handling, check that inputs are in the correct range
    check_values(vals, metadata)


    #if not (np.all(vals >= min_params) and np.all(max_params >= vals)):
    #    min_problem = np.where(vals < min_params)[0]
    #    max_problem = np.where(max_params < vals)[0]
    #
    #    out_str = "Some parameters input are not in the correct range:\n"
    #    for i in min_problem:
    #        out_str = out_str + params[i] + " : " + str(vals[i]) + " < min_trained_value = " + str(min_params[i]) + "\n"
    #    for i in max_problem:
    #        out_str = out_str + params[i] + " : " + str(vals[i]) + " > max_trained_value = " + str(max_params[i]) + "\n"
    #
    #    out_str = out_str.strip('\n')
    #    raise ValueError(out_str)

    return vals


def uniform_input_values(metadata: MetaData, default:dict = DEFAULT_VALUES, **kwargs):
    vals = input_values(metadata, default, **kwargs)
    return true_to_uniform(vals, metadata.parameters_min_val, metadata.parameters_max_val)

def uniform_input_array(theta : np.ndarray, metadata: MetaData):
    check_values(theta, metadata)
    return true_to_uniform(theta, metadata.parameters_min_val, metadata.parameters_max_val)

# ---------------------------------------------------

# ---------------------------------------------------
# PREDICTION FUNCTIONS


def predict_classifier(classifier: Classifier | None = None, 
                       default : dict | None = None, 
                       **kwargs) -> bool:
    """
    Prediction of the classifier. See `predict_classifier_numpy` for a more
    advanced but faster method (working with numpy arrays).

    Parameters
    -----------
    classifier: nnero.Classifier | None, optional
        Classifier object already trained. Default is None. If None, the default
        classifier `DefaultClassifier` is used.
    **kwargs
        Any value for a parameter the classifier has been trained on.
        Keys should corresponds to `classifier.parameters_name`.

    Returns
    --------
    bool
        True for an early reionization.
        False for a late reionization.
    """

    if default is None:
        default = DEFAULT_VALUES
    
    # if no classifier pass as input, load the default one
    if classifier is None:
        classifier = Classifier.load()

    u_vals = uniform_input_values(classifier.metadata, default, **kwargs)
    u_vals = torch.tensor(u_vals, dtype=torch.float32)
    
    with torch.no_grad():
        res = classifier.forward(u_vals)
        res = np.rint(res.numpy())
        res = res.astype(bool)

    return res[0]


def predict_classifier_numpy(theta: np.ndarray, 
                             classifier: Classifier | None = None) -> np.ndarray:
    """
    Prediction of the classifier. Same as `predict_classifier` but much faster
    as it work with numpy arrays.

    Parameters
    -----------
    theta: np.ndarray
        Array of shape (n, p) of parameters, where p is the number of parameters.
        The p parameters values should be given in the order they were fed to the
        classifier for training. These parameters and their order are accesible 
        calling `classifier.info()`.
    classifier: nnero.Classifier | None, optional
        Classifier object already trained. Default is None. If None, the default
        classifier `DefaultClassifier` is used.

    Returns
    --------
    np.ndarray
        Array of booleans.
        True for an early reionization.
        False for a late reionization.
    """

    if classifier is None:
        classifier = Classifier.load()

    u_vals = uniform_input_array(theta, classifier.metadata)
    u_vals = torch.tensor(u_vals, dtype=torch.float32)
    
    with torch.no_grad():
        res = classifier.forward(u_vals)
        res = np.rint(res.numpy())
        res = res.astype(bool)

    return res



def predict_regressor(regressor: Regressor | None = None, 
                      default: dict | None = None, 
                      **kwargs) -> np.ndarray:
    """
    Prediction of the regressor. See `predict_regressor_numpy` for a more
    advanced but faster method (working with numpy arrays).

    Parameters
    -----------
    regressor: nnero.Regressor | None, optional
        Regressor object already trained. Default is None. If None, the default
        regressor `DefaultRegressor` is used.
    default : dict | None, optional
        Dictionnary of default values, by default None.
        The default dictionnaly should have keys corresponding to those of the trained dataset
        as accessible from `regressor.info()`.
    **kwargs
        Any value for a parameter the regressor has been trained on.
        Keys should corresponds to `regressor.parameters_name`.

    Returns
    --------
    np.ndarray
        Array of values for X_e.
    """

    if default is None:
        default = DEFAULT_VALUES

    # if no regressor passed as input, load the default one
    if regressor is None:
        regressor = Regressor.load()
     
    u_vals = uniform_input_values(regressor.metadata, default, **kwargs)
    u_vals = torch.tensor(u_vals, dtype=torch.float32)
    
    with torch.no_grad():
        res = regressor.forward(u_vals).numpy()

    return res




def predict_regressor_numpy(theta: np.ndarray, 
                            regressor: Regressor | None = None) -> np.ndarray:
    """
    Prediction of the regressor. Same as `predict_regressor` but much faster
    as it work with numpy arrays.

    Parameters
    -----------
    theta: np.ndarray
        Array of shape (n, p) of parameters, where p is the number of parameters.
        The p parameters values should be given in the order they were fed to the
        regressor for training. These parameters and their order are accesible 
        calling `regressor.info()`.
    regressor: nnero.Regressor | None, optional
        Regressor object already trained. Default is None. If None, the default
        regressor `DefaultRegressor` is used.

    Returns
    --------
    np.ndarray
        2D array of shape (n, q) for the values for X_e
        wit q the number of redshift bins. 
    """

    if regressor is None:
        regressor = Regressor.load()

    u_vals = uniform_input_array(theta, regressor.metadata)
    u_vals = torch.tensor(u_vals, dtype=torch.float32)
    
    with torch.no_grad():
        res = regressor.forward(u_vals).numpy()

    return res   



def predict_interpolator(interpolator: Interpolator | None = None, 
                         default: dict | None = None, 
                         parameter: str | None = None,
                         **kwargs) -> np.ndarray:
    """
    Prediction of the interpolator. See `predict_interpolator_numpy` for a more
    advanced but faster method (working with numpy arrays).

    Parameters
    -----------
    interpolator: nnero.Interpolator | None, optional
        Interpolator object already trained. Default is None. If None, the default
        interpolator `DefaultInterpolator_<parameter>` is used.
    default : dict | None, optional
        Dictionnary of default values, by default None.
        The default dictionnaly should have keys corresponding to those of the trained dataset
        as accessible from `interpolator.info()`.
    parameter: str | None, optional
        If no interpolator given in input, must be specified to know which interpolator to load.
        Default is None.
    **kwargs
        Any value for a parameter the interpolator has been trained on.
        Keys should corresponds to `interpolator.parameters_name`.

    Returns
    --------
    float
        Interpolated value.

    Raises
    ------
    ValueError
        If no interpolator given, need to specify on which parameter we cant to interpolate
        in order to load the associated default interpolator.
    """

    if default is None:
        default = DEFAULT_VALUES

    # if no interpolator passed as input, load the default one
    if interpolator is None:
        if parameter is not None:
            interpolator = Interpolator.load("DefaultInterpolator_" + parameter)
        else:
            raise ValueError("Need to pass the parameter to interpolate to the predictor.")
     
    u_vals = uniform_input_values(interpolator.metadata, default, **kwargs)
    u_vals = torch.tensor(u_vals, dtype=torch.float32)
    
    with torch.no_grad():
        res = interpolator.forward(u_vals).numpy()

    return res




def predict_interpolator_numpy(theta: np.ndarray, 
                               interpolator: Regressor | None = None,
                               parameter: str | None = None) -> np.ndarray:
    """
    Prediction of the interpolator. Same as `predict_interpolator` but much faster
    as it work with numpy arrays.

    Parameters
    -----------
    theta: np.ndarray
        Array of shape (n, p) of parameters, where p is the number of parameters.
        The p parameters values should be given in the order they were fed to the
        regressor for training. These parameters and their order are accesible 
        calling `regressor.info()`.
    interpolator: nnero.Interpolator | None, optional
        Regressor object already trained. Default is None. If None, the default
        regressor `DefaultRegressor` is used.
    parameter: str | None, optional
        If no interpolator given in input, must be specified to know which interpolator to load.
        Default is None.

    Returns
    --------
    np.ndarray
        1D array of shape n for the inteprolated values.

    Raises
    ------
    ValueError
        If no interpolator given, need to specify on which parameter we cant to interpolate
        in order to load the associated default interpolator.
    """

    # if no interpolator passed as input, load the default one
    if interpolator is None:
        if parameter is not None:
            interpolator = Interpolator.load("DefaultInterpolator_" + parameter)
        else:
            raise ValueError("Need to pass the parameter to interpolate to the predictor.")
        
    u_vals = uniform_input_array(theta, interpolator.metadata)
    u_vals = torch.tensor(u_vals, dtype=torch.float32)
    
    with torch.no_grad():
        res = interpolator.forward(u_vals).numpy()

    return res   





def predict_Xe(classifier: Classifier | None = None, 
               regressor:  Regressor  | None = None, 
               default: dict | None = None,
               **kwargs) -> np.ndarray | bool :
    """
    Prediction of the free electron fraction X_e, taking into account
    the selection from the classifier. See `predict_Xe_numpy` for a more
    advanced but faster method (working with numpy arrays).

    Parameters
    ----------
    classifier : Classifier | None, optional
        Classifier object already trained. Default is None. If None, the default
        classifier `DefaultClassifier` is used.
    regressor : Regressor | None, optional
        Regressor object already trained. Default is None. If None, the default
        regressor `DefaultRegressor` is used.
    default : dict | None, optional
        Dictionnary of default values, by default None.
        The default dictionnaly should have keys corresponding to those of the trained dataset
        as accessible from `regressor.info()` or `classifier.info()`.
    **kwargs
        Any value for a parameter the classifier and regressor have been trained on.
        Keys should corresponds to `regressor.parameters_name` or `classifier.parameters_name`.
        
    Returns
    -------
    np.ndarray | bool
        Returns `False` if the classifier returns `False`. Returns the prediction
        from the regressor for the free electron fraction X_e otherwise.
    """
    if default is None:
        default = DEFAULT_VALUES
    
    early = predict_classifier(classifier, default, **kwargs)

    if not early:
        return False
    
    xHII = predict_regressor(regressor, default, **kwargs)
    return xHII


def predict_Xe_numpy(theta: np.ndarray,
                     classifier: Classifier | None = None, 
                     regressor:  Regressor  | None = None) -> np.ndarray:
    """
    Prediction of the free electron fraction X_e, taking into account
    the selection from the classifier. Same as `predict_Xe` but much faster
    as it work with numpy arrays.

    Parameters
    ----------
    theta: np.ndarray
        Array of shape (n, p) of parameters, where p is the number of parameters.
        The p parameters values should be given in the order they were fed to the
        regressor for training. These parameters and their order are accesible 
        calling `regressor.info()`.
    classifier : Classifier | None, optional
        Classifier object already trained. Default is None. If None, the default
        classifier `DefaultClassifier` is used.
    regressor : Regressor | None, optional
        Regressor object already trained. Default is None. If None, the default
        regressor `DefaultRegressor` is used.

    Returns
    -------
    np.ndarray
        2D array of shape (n, q) for the values for X_e
        with q the number for parameters.
        When the classifier outputs false, the array is filled with -1. 
    """
    
    if regressor is None:
        regressor = Regressor.load()
    
    mask = predict_classifier_numpy(theta, classifier)
    xHII = -np.ones((mask.shape[0], len(regressor.metadata.z)), dtype=np.float64)

    xHII[mask, :] = predict_regressor_numpy(theta[mask, :], regressor)

    return xHII





def predict_parameter(classifier: Classifier | None = None,  
                      interpolator:  Interpolator | None = None,  
                      default: dict | None = None,
                      parameter: str | None = None,
                      **kwargs) -> np.ndarray | bool :
    """
    Prediction of the interpolated parameter, taking into account
    the selection from the classifier. See `predict_parameter_numpy` for a more
    advanced but faster method (working with numpy arrays).

    Parameters
    ----------
    classifier : Classifier | None, optional
        Classifier object already trained. Default is None. If None, the default
        classifier `DefaultClassifier` is used.
    interpolator : Interpolator | None, optional
        Interpolator object already trained. Default is None. If None, the default
        interpolator `DefaultInterpolator_<parameter>` is used.
    default : dict | None, optional
        Dictionnary of default values, by default None.
        The default dictionnaly should have keys corresponding to those of the trained dataset
        as accessible from `interpolator.info()` or `classifier.info()`.
    **kwargs
        Any value for a parameter the classifier and interpolator have been trained on.
        Keys should corresponds to `regressor.parameters_name` or `classifier.parameters_name`.
        
    Returns
    -------
    np.ndarray | bool
        Returns `False` if the classifier returns `False`. Returns the prediction
        from the interpolator for the parameter otherwise.
    """

    if default is None:
        default = DEFAULT_VALUES
    
    early = predict_classifier(classifier, default, **kwargs)

    if not early:
        return False
    
    return predict_interpolator(interpolator, default, parameter, **kwargs)


def predict_parameter_numpy(theta: np.ndarray,
                            classifier: Classifier | None = None, 
                            interpolator:  Interpolator  | None = None,
                            parameter: str | None = None) -> np.ndarray:
    """
    Prediction of the interpolated parameter, taking into account
    the selection from the classifier. Same as `predict_parameter` but much faster
    as it work with numpy arrays.

    Parameters
    ----------
    theta: np.ndarray
        Array of shape (n, p) of parameters, where p is the number of parameters.
        The p parameters values should be given in the order they were fed to the
        regressor for training. These parameters and their order are accesible 
        calling `regressor.info()`.
    classifier : Classifier | None, optional
        Classifier object already trained. Default is None. If None, the default
        classifier `DefaultClassifier` is used.
    interpolator : Interpolator | None, optional
        Interpolator object already trained. Default is None. If None, the default
        interpolator `DefaultInterpolator` is used.

    Returns
    -------
    np.ndarray
        1D array of shape (n) for the value of the parameter
        When the classifier outputs false, the array is filled with -1. 
    """
    
    # if no interpolator passed as input, load the default one
    if interpolator is None:
        if parameter is not None:
            interpolator = Interpolator.load("DefaultInterpolator_" + parameter)
        else:
            raise ValueError("Need to pass the parameter to interpolate to the predictor.")
    
    mask = predict_classifier_numpy(theta, classifier)
    res = -np.ones((mask.shape[0]), dtype=np.float64)

    res[mask] = predict_interpolator_numpy(theta[mask, :], interpolator, parameter)

    return res




def predict_tau_from_Xe(xe: np.ndarray, 
                        metadata: MetaData, 
                        default: dict | None = None, 
                        **kwargs) -> float:
    """
    Predict the optical depth to reionization from an array of X_e.
    See `predict_tau_from_Xe_numpy` for a more advanced but faster method 
    (working with numpy arrays).

    Parameters
    ----------
    xe : np.ndarray
        Array for the free electron fraction X_e = xHII(1+db).
    metadata : MetaData
        Metadata object (attached to the networks) describing properties
        of the data the networks have been trained on.
    default : dict | None, optional
        Dictionnary of default values, by default None.
        The default dictionnaly should have keys corresponding to those of the trained dataset
        as accessible from `metadata.parameters_name`.
    **kwargs
        Any value for a parameter in metadata.
        Keys should corresponds to `metadata.parameters_name`.
        
    Returns
    -------
    float
    """

    if default is None:
        default = DEFAULT_VALUES

    vals = input_values(metadata, default, **kwargs)
    
    omega_b = np.array([vals[metadata.pos_omega_b]])[None, :]
    omega_c = np.array([vals[metadata.pos_omega_dm]])[None, :]
    hlittle = np.array([vals[metadata.pos_hlittle]])[None, :]
    
    return optical_depth_no_rad(metadata.z[None, :], xe[None, :], omega_b, omega_c, hlittle)[0]


def predict_tau_from_Xe_numpy(xe: np.ndarray, 
                              theta : np.ndarray, 
                              metadata : MetaData) -> np.ndarray:
    """
    Predict the optical depth to reionization from an array of X_e.
    Same as `predict_tau_from_Xe` but much faster as it work with numpy arrays.

    Parameters
    ----------
    xe : np.ndarray
        Array for the free electron fraction X_e = xHII(1+db).
    theta: np.ndarray
        Array of shape (n, p) of parameters, where p is the number of parameters.
        The p parameters values should be given in the order they were fed to the
        regressor for training. These parameters and their order are accesible 
        calling `metadata.parameters_name`.
    metadata : MetaData
        Metadata object (attached to the networks) describing properties
        of the data the networks have been trained on.

    Returns
    -------
    np.ndarray
        Array of values of the optical depth to reionization.
    """
    
    omega_b = theta[:, metadata.pos_omega_b]
    omega_c = theta[:, metadata.pos_omega_dm]
    hlittle = theta[:, metadata.pos_hlittle]

    return optical_depth_no_rad(metadata.z[None, :], xe, omega_b, omega_c, hlittle)



def predict_tau(classifier: Classifier | None = None,
                regressor: Regressor   | None = None,
                default: dict | None = None,
                **kwargs) -> float:
    """
    Predict the optical depth to reionization from a trained classifier and regressor
    as well as parameters passed in kwargs or default. See `predict_tau_numpy` 
    for a more advanced but faster method (working with numpy arrays).

    Parameters
    ----------
    classifier : Classifier | None, optional
        Classifier object already trained. Default is None. If None, the default
        classifier `DefaultClassifier` is used.
    regressor : Regressor | None, optional
        Regressor object already trained. Default is None. If None, the default
        regressor `DefaultRegressor` is used.
    default : dict | None, optional
        Dictionnary of default values, by default None.
        The default dictionnaly should have keys corresponding to those of the trained dataset
        as accessible from `regressor.info()` or `classifier.info()`.
    **kwargs
        Any value for a parameter the classifier and regressor have been trained on.
        Keys should corresponds to `regressor.parameters_name` or `classifier.parameters_name`.
        

    Returns
    -------
    float
    """
    
    if default is None:
        default = DEFAULT_VALUES

    if regressor is None:
        regressor = Regressor.load()

    xHII = predict_Xe(classifier, regressor, default, **kwargs)
    
    if xHII is False:
        return -1
    
    return predict_tau_from_Xe(xHII, regressor.metadata, default, **kwargs)



def predict_tau_numpy(theta: np.ndarray, 
                      classifier: Classifier | None = None,
                      regressor: Regressor   | None = None) -> np.ndarray:
    """
    Predict the optical depth to reionization from a trained classifier and regressor
    as well as parameters passed in kwargs or default. Same as `predict_tau` but much faster
    as it work with numpy arrays.

    Parameters
    ----------
    theta: np.ndarray
        Array of shape (n, p) of parameters, where p is the number of parameters.
        The p parameters values should be given in the order they were fed to the
        regressor for training. These parameters and their order are accesible 
        calling `regressor.info()` or `classifier.info()`.
    classifier : Classifier | None, optional
        Classifier object already trained. Default is None. If None, the default
        classifier `DefaultClassifier` is used.
    regressor : Regressor | None, optional
        Regressor object already trained. Default is None. If None, the default
        regressor `DefaultRegressor` is used.

    Returns
    -------
    np.ndarray
        Array of values of the optical depth to reionization.
    """
    
    if regressor is None:
        regressor = Regressor.load()
    
    xHII = predict_Xe_numpy(theta, classifier, regressor)
    
    res  = - np.ones(len(xHII))
    mask = (xHII[:, -1] != -1)
    
    res[mask]  = predict_tau_from_Xe_numpy(xHII[mask, :], theta[mask, :], regressor.metadata)

    return res

###########################################

# Defining equivalent names for functions that are now deprecated

def predict_xHII(classifier: Classifier | None = None, 
                 regressor:  Regressor  | None = None, 
                 default: dict | None = None,
                 **kwargs) -> np.ndarray:
    """
    .. deprecated::
        See `predict_Xe` instead.
    """
    return predict_Xe(classifier, regressor, default, **kwargs)

def predict_xHII_numpy(theta: np.ndarray,
                     classifier: Classifier | None = None, 
                     regressor:  Regressor  | None = None) -> np.ndarray:
    """
    .. deprecated::
        See `predict_Xe_numpy` instead.
    """
    return predict_Xe_numpy(theta, classifier, regressor)

def predict_tau_from_xHII(Xe, metadata : MetaData, default: dict | None = None, **kwargs) -> float:
    """
    .. deprecated::
        See `predict_tau_from_Xe` instead.
    """
    return predict_tau_from_Xe(Xe, metadata, default, **kwargs)

def predict_tau_from_xHII_numpy(Xe, theta : np.ndarray, metadata : MetaData) -> np.ndarray:
    """
    .. deprecated::
        See `predict_tau_from_Xe_numpy` instead.
    """
    return predict_tau_from_Xe_numpy(Xe, theta, metadata)



import pytest
import nnero

def test_load_classifier():
    assert nnero.Classifier.load()

def test_load_regressor():
    assert nnero.Regressor.load()

def test_predict_classifier():
    assert (nnero.predict_classifier() is not None)

def test_predict_regressor():
    assert (nnero.predict_Xe() is not None)
